# Copyright (C) 2008-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2007 Canonical Ltd
# Copyright (C) 2008 John Carr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Converters, etc for going between Bazaar and Git ids."""

import base64
import stat
from typing import Optional

import fastbencode as bencode

from .. import errors, foreign, trace, urlutils
from ..foreign import ForeignRevision, ForeignVcs, VcsMappingRegistry
from ..revision import NULL_REVISION, Revision
from .errors import NoPushSupport
from .hg import extract_hg_metadata, format_hg_metadata
from .roundtrip import (CommitSupplement, extract_bzr_metadata,
                        inject_bzr_metadata)

DEFAULT_FILE_MODE = stat.S_IFREG | 0o644
HG_RENAME_SOURCE = b"HG:rename-source"
HG_EXTRA = b"HG:extra"

# This HG extra is used to indicate the commit that this commit was based on.
HG_EXTRA_AMEND_SOURCE = b"amend_source"
HG_EXTRA_REBASE_SOURCE = b"rebase_source"
HG_EXTRA_ABSORB_SOURCE = b"absorb_source"
HG_EXTRA_SOURCE = b"source"
HG_EXTRA_INTERMEDIATE_SOURCE = b"intermediate-source"
HG_EXTRA_TOPIC = b"topic"
HG_EXTRA_REWRITE_NOISE = b"_rewrite_noise"

FILE_ID_PREFIX = b'git:'

# Always the same.
ROOT_ID = b"TREE_ROOT"


class UnknownCommitExtra(errors.BzrError):
    _fmt = "Unknown extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        errors.BzrError.__init__(self)
        self.object = object
        self.fields = ",".join(fields)


class UnknownMercurialCommitExtra(errors.BzrError):
    _fmt = "Unknown mercurial extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        errors.BzrError.__init__(self)
        self.object = object
        self.fields = b",".join(fields)


class UnknownCommitEncoding(errors.BzrError):
    _fmt = "Unknown commit encoding: %(encoding)s"

    def __init__(self, encoding):
        errors.BzrError.__init__(self)
        self.encoding = encoding


def escape_file_id(file_id):
    file_id = file_id.replace(b'_', b'__')
    file_id = file_id.replace(b' ', b'_s')
    file_id = file_id.replace(b'\x0c', b'_c')
    return file_id


def unescape_file_id(file_id):
    ret = bytearray()
    i = 0
    while i < len(file_id):
        if file_id[i:i + 1] != b'_':
            ret.append(file_id[i])
        else:
            if file_id[i + 1:i + 2] == b'_':
                ret.append(b"_"[0])
            elif file_id[i + 1:i + 2] == b's':
                ret.append(b" "[0])
            elif file_id[i + 1:i + 2] == b'c':
                ret.append(b"\x0c"[0])
            else:
                raise ValueError("unknown escape character %s" %
                                 file_id[i + 1:i + 2])
            i += 1
        i += 1
    return bytes(ret)


def fix_person_identifier(text):
    if b"<" not in text and b">" not in text:
        username = text
        email = text
    elif b">" not in text:
        return text + b">"
    else:
        if text.rindex(b">") < text.rindex(b"<"):
            raise ValueError(text)
        username, email = text.split(b"<", 2)[-2:]
        email = email.split(b">", 1)[0]
        if username.endswith(b" "):
            username = username[:-1]
    return b"%s <%s>" % (username, email)


def decode_git_path(path):
    """Take a git path and decode it."""
    return path.decode('utf-8', 'surrogateescape')


def encode_git_path(path):
    """Take a regular path and encode it for git."""
    return path.encode('utf-8', 'surrogateescape')


def warn_escaped(commit, num_escaped):
    trace.warning("Escaped %d XML-invalid characters in %s. Will be unable "
                  "to regenerate the SHA map.", num_escaped, commit)


def warn_unusual_mode(commit, path, mode):
    trace.mutter("Unusual file mode %o for %s in %s. Storing as revision "
                 "property. ", mode, path, commit)


class BzrGitMapping(foreign.VcsMapping):
    """Class that maps between Git and Bazaar semantics."""
    experimental = False

    BZR_DUMMY_FILE = None  # type: Optional[str]

    def is_special_file(self, filename):
        return (filename in (self.BZR_DUMMY_FILE, ))

    def __init__(self):
        super().__init__(foreign_vcs_git)

    def __eq__(self, other):
        return (type(self) is type(other)
                and self.revid_prefix == other.revid_prefix)

    @classmethod
    def revision_id_foreign_to_bzr(cls, git_rev_id):
        """Convert a git revision id handle to a Bazaar revision id."""
        from dulwich.protocol import ZERO_SHA
        if git_rev_id == ZERO_SHA:
            return NULL_REVISION
        return b"%s:%s" % (cls.revid_prefix, git_rev_id)

    @classmethod
    def revision_id_bzr_to_foreign(cls, bzr_rev_id):
        """Convert a Bazaar revision id to a git revision id handle."""
        if not bzr_rev_id.startswith(b"%s:" % cls.revid_prefix):
            raise errors.InvalidRevisionId(bzr_rev_id, cls)
        return bzr_rev_id[len(cls.revid_prefix) + 1:], cls()

    def generate_file_id(self, path):
        # Git paths are just bytestrings
        # We must just hope they are valid UTF-8..
        if isinstance(path, str):
            path = encode_git_path(path)
        if path == b"":
            return ROOT_ID
        return FILE_ID_PREFIX + escape_file_id(path)

    def parse_file_id(self, file_id):
        if file_id == ROOT_ID:
            return ""
        if not file_id.startswith(FILE_ID_PREFIX):
            raise ValueError
        return decode_git_path(unescape_file_id(file_id[len(FILE_ID_PREFIX):]))

    def import_unusual_file_modes(self, rev, unusual_file_modes):
        if unusual_file_modes:
            ret = [(path, unusual_file_modes[path])
                   for path in sorted(unusual_file_modes.keys())]
            rev.properties['file-modes'] = bencode.bencode(ret)

    def export_unusual_file_modes(self, rev):
        try:
            file_modes = rev.properties['file-modes']
        except KeyError:
            return {}
        else:
            return dict(bencode.bdecode(file_modes.encode("utf-8")))

    def _generate_git_svn_metadata(self, rev, encoding):
        try:
            git_svn_id = rev.properties["git-svn-id"]
        except KeyError:
            return ""
        else:
            return "\ngit-svn-id: %s\n" % git_svn_id.encode(encoding)

    def _generate_hg_message_tail(self, rev):
        extra = {}
        renames = []
        branch = 'default'
        for name in rev.properties:
            if name == 'hg:extra:branch':
                branch = rev.properties['hg:extra:branch']
            elif name.startswith('hg:extra'):
                extra[name[len('hg:extra:'):]] = base64.b64decode(
                    rev.properties[name])
            elif name == 'hg:renames':
                renames = bencode.bdecode(base64.b64decode(
                    rev.properties['hg:renames']))
            # TODO: Export other properties as 'bzr:' extras?
        ret = format_hg_metadata(renames, branch, extra)
        if not isinstance(ret, bytes):
            raise TypeError(ret)
        return ret

    def _extract_git_svn_metadata(self, rev, message):
        lines = message.split("\n")
        if not (lines[-1] == "" and len(lines) >= 2 and
                lines[-2].startswith("git-svn-id:")):
            return message
        git_svn_id = lines[-2].split(": ", 1)[1]
        rev.properties['git-svn-id'] = git_svn_id
        (url, rev, uuid) = parse_git_svn_id(git_svn_id)
        # FIXME: Convert this to converted-from property somehow..
        return "\n".join(lines[:-2])

    def _extract_hg_metadata(self, rev, message):
        (message, renames, branch, extra) = extract_hg_metadata(message)
        if branch is not None:
            rev.properties['hg:extra:branch'] = branch
        for name, value in extra.items():
            rev.properties['hg:extra:' + name] = base64.b64encode(value)
        if renames:
            rev.properties['hg:renames'] = base64.b64encode(bencode.bencode(
                [(new, old) for (old, new) in renames.items()]))
        return message

    def _extract_bzr_metadata(self, rev, message):
        (message, metadata) = extract_bzr_metadata(message)
        return message, metadata

    def _decode_commit_message(self, rev, message, encoding):
        if message is None:
            decoded_message = None
        else:
            decoded_message = message.decode(encoding)
        return decoded_message, CommitSupplement()

    def _encode_commit_message(self, rev, message, encoding):
        if message is None:
            return None
        else:
            return message.encode(encoding)

    def export_commit(self, rev, tree_sha, parent_lookup, lossy,
                      verifiers):
        """Turn a Bazaar revision in to a Git commit

        :param tree_sha: Tree sha for the commit
        :param parent_lookup: Function for looking up the GIT sha equiv of a
            bzr revision
        :param lossy: Whether to store roundtripping information.
        :param verifiers: Verifiers info
        :return dulwich.objects.Commit represent the revision:
        """
        from dulwich.objects import Commit, Tag
        commit = Commit()
        commit.tree = tree_sha
        if not lossy:
            metadata = CommitSupplement()
            metadata.verifiers = verifiers
        else:
            metadata = None
        parents = []
        for p in rev.parent_ids:
            try:
                git_p = parent_lookup(p)
            except KeyError:
                git_p = None
                if metadata is not None:
                    metadata.explicit_parent_ids = rev.parent_ids
            if git_p is not None:
                if len(git_p) != 40:
                    raise AssertionError("unexpected length for %r" % git_p)
                parents.append(git_p)
        commit.parents = parents
        try:
            encoding = rev.properties['git-explicit-encoding']
        except KeyError:
            encoding = rev.properties.get('git-implicit-encoding', 'utf-8')
        try:
            commit.encoding = rev.properties['git-explicit-encoding'].encode(
                'ascii')
        except KeyError:
            pass
        commit.committer = fix_person_identifier(rev.committer.encode(
            encoding))
        first_author = rev.get_apparent_authors()[0]
        if ',' in first_author and first_author.count('>') > 1:
            first_author = first_author.split(',')[0]
        commit.author = fix_person_identifier(
            first_author.encode(encoding))
        # TODO(jelmer): Don't use this hack.
        long = getattr(__builtins__, 'long', int)
        commit.commit_time = long(rev.timestamp)
        if 'author-timestamp' in rev.properties:
            commit.author_time = long(rev.properties['author-timestamp'])
        else:
            commit.author_time = commit.commit_time
        commit._commit_timezone_neg_utc = (
            "commit-timezone-neg-utc" in rev.properties)
        commit.commit_timezone = rev.timezone
        commit._author_timezone_neg_utc = (
            "author-timezone-neg-utc" in rev.properties)
        if 'author-timezone' in rev.properties:
            commit.author_timezone = int(rev.properties['author-timezone'])
        else:
            commit.author_timezone = commit.commit_timezone
        if 'git-gpg-signature' in rev.properties:
            commit.gpgsig = rev.properties['git-gpg-signature'].encode(
                'utf-8', 'surrogateescape')
        commit.message = self._encode_commit_message(rev, rev.message,
                                                     encoding)
        if not isinstance(commit.message, bytes):
            raise TypeError(commit.message)
        if metadata is not None:
            try:
                mapping_registry.parse_revision_id(rev.revision_id)
            except errors.InvalidRevisionId:
                metadata.revision_id = rev.revision_id
            mapping_properties = {
                'author', 'author-timezone', 'author-timezone-neg-utc',
                'commit-timezone-neg-utc', 'git-implicit-encoding',
                'git-gpg-signature', 'git-explicit-encoding',
                'author-timestamp', 'file-modes'}
            for k, v in rev.properties.items():
                if k not in mapping_properties:
                    metadata.properties[k] = v
        if not lossy and metadata:
            if self.roundtripping:
                commit.message = inject_bzr_metadata(commit.message, metadata,
                                                     encoding)
            else:
                raise NoPushSupport(
                    None, None, self, revision_id=rev.revision_id)
        if not isinstance(commit.message, bytes):
            raise TypeError(commit.message)
        i = 0
        propname = 'git-mergetag-0'
        while propname in rev.properties:
            commit.mergetag.append(
                Tag.from_string(rev.properties[propname].encode('utf-8', 'surrogateescape')))
            i += 1
            propname = 'git-mergetag-%d' % i
        try:
            extra = commit._extra
        except AttributeError:
            extra = commit.extra
        if 'git-extra' in rev.properties:
            for l in rev.properties['git-extra'].splitlines():
                (k, v) = l.split(' ', 1)
                extra.append(
                    (k.encode('utf-8', 'surrogateescape'),
                     v.encode('utf-8', 'surrogateescape')))
        return commit

    def get_revision_id(self, commit):
        if commit.encoding:
            encoding = commit.encoding.decode('ascii')
        else:
            encoding = 'utf-8'
        if commit.message is not None:
            try:
                message, metadata = self._decode_commit_message(
                    None, commit.message, encoding)
            except UnicodeDecodeError:
                pass
            else:
                if metadata.revision_id:
                    return metadata.revision_id
        return self.revision_id_foreign_to_bzr(commit.id)

    def import_commit(self, commit, lookup_parent_revid, strict=True):
        """Convert a git commit to a bzr revision.

        :return: a `breezy.revision.Revision` object, foreign revid and a
            testament sha1
        """
        if commit is None:
            raise AssertionError("Commit object can't be None")
        rev = ForeignRevision(commit.id, self,
                              self.revision_id_foreign_to_bzr(commit.id))
        rev.git_metadata = None

        def decode_using_encoding(rev, commit, encoding):
            try:
                rev.committer = commit.committer.decode(encoding)
            except LookupError:
                raise UnknownCommitEncoding(encoding)
            try:
                if commit.committer != commit.author:
                    rev.properties['author'] = commit.author.decode(encoding)
            except LookupError:
                raise UnknownCommitEncoding(encoding)
            rev.message, rev.git_metadata = self._decode_commit_message(
                rev, commit.message, encoding)

        if commit.encoding is not None:
            rev.properties['git-explicit-encoding'] = commit.encoding.decode(
                'ascii')
        if commit.encoding is not None and commit.encoding != b'false':
            decode_using_encoding(rev, commit, commit.encoding.decode('ascii'))
        else:
            for encoding in ('utf-8', 'latin1'):
                try:
                    decode_using_encoding(rev, commit, encoding)
                except UnicodeDecodeError:
                    pass
                else:
                    if encoding != 'utf-8':
                        rev.properties['git-implicit-encoding'] = encoding
                    break
        if commit.commit_time != commit.author_time:
            rev.properties['author-timestamp'] = str(commit.author_time)
        if commit.commit_timezone != commit.author_timezone:
            rev.properties['author-timezone'] = "%d" % commit.author_timezone
        if commit._author_timezone_neg_utc:
            rev.properties['author-timezone-neg-utc'] = ""
        if commit._commit_timezone_neg_utc:
            rev.properties['commit-timezone-neg-utc'] = ""
        if commit.gpgsig:
            rev.properties['git-gpg-signature'] = commit.gpgsig.decode(
                'utf-8', 'surrogateescape')
        if commit.mergetag:
            for i, tag in enumerate(commit.mergetag):
                rev.properties['git-mergetag-%d' % i] = tag.as_raw_string().decode(
                    'utf-8', 'surrogateescape')
        rev.timestamp = commit.commit_time
        rev.timezone = commit.commit_timezone
        rev.parent_ids = None
        if rev.git_metadata is not None:
            md = rev.git_metadata
            roundtrip_revid = md.revision_id
            if md.explicit_parent_ids:
                rev.parent_ids = md.explicit_parent_ids
            rev.properties.update(md.properties)
            verifiers = md.verifiers
        else:
            roundtrip_revid = None
            verifiers = {}
        if rev.parent_ids is None:
            parents = []
            for p in commit.parents:
                try:
                    parents.append(lookup_parent_revid(p))
                except KeyError:
                    parents.append(self.revision_id_foreign_to_bzr(p))
            rev.parent_ids = list(parents)
        unknown_extra_fields = []
        extra_lines = []
        try:
            extra = commit._extra
        except AttributeError:
            extra = commit.extra
        for k, v in extra:
            if k == HG_RENAME_SOURCE:
                extra_lines.append(
                    k.decode('utf-8', 'surrogateescape') + ' ' +
                    v.decode('utf-8', 'surrogateescape') + '\n')
            elif k == HG_EXTRA:
                hgk, hgv = v.split(b':', 1)
                if hgk not in (HG_EXTRA_AMEND_SOURCE, HG_EXTRA_REBASE_SOURCE,
                               HG_EXTRA_ABSORB_SOURCE, HG_EXTRA_INTERMEDIATE_SOURCE,
                               HG_EXTRA_SOURCE, HG_EXTRA_TOPIC,
                               HG_EXTRA_REWRITE_NOISE) and strict:
                    raise UnknownMercurialCommitExtra(commit, [hgk])
                extra_lines.append(
                    k.decode('utf-8', 'surrogateescape') + ' ' +
                    v.decode('utf-8', 'surrogateescape') + '\n')
            else:
                unknown_extra_fields.append(k)
        if unknown_extra_fields and strict:
            raise UnknownCommitExtra(
                commit,
                [f.decode('ascii', 'replace') for f in unknown_extra_fields])
        if extra_lines:
            rev.properties['git-extra'] = ''.join(extra_lines)
        return rev, roundtrip_revid, verifiers


class BzrGitMappingv1(BzrGitMapping):
    revid_prefix = b'git-v1'
    experimental = False

    def __str__(self):
        return self.revid_prefix.decode('utf-8')


class BzrGitMappingExperimental(BzrGitMappingv1):
    revid_prefix = b'git-experimental'
    experimental = True
    roundtripping = False

    BZR_DUMMY_FILE = '.bzrdummy'

    def _decode_commit_message(self, rev, message, encoding):
        if rev is None:
            rev = Revision()
        message = self._extract_hg_metadata(rev, message)
        message = self._extract_git_svn_metadata(rev, message)
        message, metadata = self._extract_bzr_metadata(rev, message)
        try:
            return message.decode(encoding), metadata
        except LookupError:
            raise UnknownCommitEncoding(encoding)

    def _encode_commit_message(self, rev, message, encoding):
        ret = message.encode(encoding)
        ret += self._generate_hg_message_tail(rev)
        ret += self._generate_git_svn_metadata(rev, encoding)
        return ret

    def import_commit(self, commit, lookup_parent_revid, strict=True):
        rev, roundtrip_revid, verifiers = super().import_commit(
                commit, lookup_parent_revid, strict)
        rev.properties['converted_revision'] = "git %s\n" % commit.id
        return rev, roundtrip_revid, verifiers


class GitMappingRegistry(VcsMappingRegistry):
    """Registry with available git mappings."""

    def revision_id_bzr_to_foreign(self, bzr_revid):
        if bzr_revid == NULL_REVISION:
            from dulwich.protocol import ZERO_SHA
            return ZERO_SHA, None
        if not bzr_revid.startswith(b"git-"):
            raise errors.InvalidRevisionId(bzr_revid, None)
        (mapping_version, git_sha) = bzr_revid.split(b":", 1)
        mapping = self.get(mapping_version)
        return mapping.revision_id_bzr_to_foreign(bzr_revid)

    parse_revision_id = revision_id_bzr_to_foreign


mapping_registry = GitMappingRegistry()
mapping_registry.register_lazy(b'git-v1', __name__,
                               "BzrGitMappingv1")
mapping_registry.register_lazy(b'git-experimental',
                               __name__, "BzrGitMappingExperimental")
# Uncomment the next line to enable the experimental bzr-git mappings.
# This will make sure all bzr metadata is pushed into git, allowing for
# full roundtripping later.
# NOTE: THIS IS EXPERIMENTAL. IT MAY EAT YOUR DATA OR CORRUPT
# YOUR BZR OR GIT REPOSITORIES. USE WITH CARE.
# mapping_registry.set_default('git-experimental')
mapping_registry.set_default(b'git-v1')


class ForeignGit(ForeignVcs):
    """The Git Stupid Content Tracker"""

    @property
    def branch_format(self):
        from .branch import LocalGitBranchFormat
        return LocalGitBranchFormat()

    @property
    def repository_format(self):
        from .repository import GitRepositoryFormat
        return GitRepositoryFormat()

    def __init__(self):
        super().__init__(mapping_registry)
        self.abbreviation = "git"

    @classmethod
    def serialize_foreign_revid(self, foreign_revid):
        return foreign_revid

    @classmethod
    def show_foreign_revid(cls, foreign_revid):
        return {"git commit": foreign_revid.decode('utf-8')}


foreign_vcs_git = ForeignGit()
default_mapping = mapping_registry.get_default()()


def symlink_to_blob(symlink_target):
    from dulwich.objects import Blob
    blob = Blob()
    if isinstance(symlink_target, str):
        symlink_target = encode_git_path(symlink_target)
    blob.data = symlink_target
    return blob


def mode_is_executable(mode):
    """Check if mode should be considered executable."""
    return bool(mode & 0o111)


def mode_kind(mode):
    """Determine the Bazaar inventory kind based on Unix file mode."""
    if mode is None:
        return None
    entry_kind = (mode & 0o700000) / 0o100000
    if entry_kind == 0:
        return 'directory'
    elif entry_kind == 1:
        file_kind = (mode & 0o70000) / 0o10000
        if file_kind == 0:
            return 'file'
        elif file_kind == 2:
            return 'symlink'
        elif file_kind == 6:
            return 'tree-reference'
        else:
            raise AssertionError(
                "Unknown file kind %d, perms=%o." % (file_kind, mode,))
    else:
        raise AssertionError(
            "Unknown kind, perms={!r}.".format(mode))


def object_mode(kind, executable):
    if kind == 'directory':
        return stat.S_IFDIR
    elif kind == 'symlink':
        mode = stat.S_IFLNK
        if executable:
            mode |= 0o111
        return mode
    elif kind == 'file':
        mode = stat.S_IFREG | 0o644
        if executable:
            mode |= 0o111
        return mode
    elif kind == 'tree-reference':
        from dulwich.objects import S_IFGITLINK
        return S_IFGITLINK
    else:
        raise AssertionError


def entry_mode(entry):
    """Determine the git file mode for an inventory entry."""
    return object_mode(entry.kind, getattr(entry, 'executable', False))


def extract_unusual_modes(rev):
    try:
        foreign_revid, mapping = mapping_registry.parse_revision_id(
            rev.revision_id)
    except errors.InvalidRevisionId:
        return {}
    else:
        return mapping.export_unusual_file_modes(rev)


def parse_git_svn_id(text):
    (head, uuid) = text.rsplit(" ", 1)
    (full_url, rev) = head.rsplit("@", 1)
    return (full_url, int(rev), uuid)


def needs_roundtripping(repo, revid):
    try:
        mapping_registry.parse_revision_id(revid)
    except errors.InvalidRevisionId:
        return True
    else:
        return False



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import errno
import re

import fastbencode as bencode

from . import errors
from .lazy_import import lazy_import

lazy_import(globals(), """
from breezy import (
    merge,
    transform,
)
from breezy.bzr import (
    pack,
    )
""")


class ShelfCorrupt(errors.BzrError):

    _fmt = "Shelf corrupt."


class NoSuchShelfId(errors.BzrError):

    _fmt = 'No changes are shelved with id "%(shelf_id)d".'

    def __init__(self, shelf_id):
        errors.BzrError.__init__(self, shelf_id=shelf_id)


class InvalidShelfId(errors.BzrError):

    _fmt = '"%(invalid_id)s" is not a valid shelf id, try a number instead.'

    def __init__(self, invalid_id):
        errors.BzrError.__init__(self, invalid_id=invalid_id)


class ShelfCreator:
    """Create a transform to shelve objects and its inverse."""

    def __init__(self, work_tree, target_tree, file_list=None):
        """Constructor.

        :param work_tree: The working tree to apply changes to. This is not
            required to be locked - a tree_write lock will be taken out.
        :param target_tree: The tree to make the working tree more similar to.
            This is not required to be locked - a read_lock will be taken out.
        :param file_list: The files to make more similar to the target.
        """
        self.work_tree = work_tree
        self.work_transform = work_tree.transform()
        try:
            self.target_tree = target_tree
            self.shelf_transform = self.target_tree.preview_transform()
            try:
                self.renames = {}
                self.creation = {}
                self.deletion = {}
                self.iter_changes = work_tree.iter_changes(
                    self.target_tree, specific_files=file_list)
            except:
                self.shelf_transform.finalize()
                raise
        except:
            self.work_transform.finalize()
            raise

    def iter_shelvable(self):
        """Iterable of tuples describing shelvable changes.

        As well as generating the tuples, this updates several members.
        Tuples may be::

           ('add file', file_id, work_kind, work_path)
           ('delete file', file_id, target_kind, target_path)
           ('rename', file_id, target_path, work_path)
           ('change kind', file_id, target_kind, work_kind, target_path)
           ('modify text', file_id)
           ('modify target', file_id, target_target, work_target)
        """
        for change in self.iter_changes:
            # don't shelve add of tree root.  Working tree should never
            # lack roots, and bzr misbehaves when they do.
            # FIXME ADHB (2009-08-09): should still shelve adds of tree roots
            # when a tree root was deleted / renamed.
            if change.kind[0] is None and change.name[1] == '':
                continue
            # Also don't shelve deletion of tree root.
            if change.kind[1] is None and change.name[0] == '':
                continue
            if change.kind[0] is None or change.versioned[0] is False:
                self.creation[change.file_id] = (
                    change.kind[1], change.name[1], change.parent_id[1], change.versioned)
                yield ('add file', change.file_id, change.kind[1], change.path[1])
            elif change.kind[1] is None or change.versioned[0] is False:
                self.deletion[change.file_id] = (
                    change.kind[0], change.name[0], change.parent_id[0], change.versioned)
                yield ('delete file', change.file_id, change.kind[0], change.path[0])
            else:
                if change.name[0] != change.name[1] or change.parent_id[0] != change.parent_id[1]:
                    self.renames[change.file_id] = (change.name, change.parent_id)
                    yield ('rename', change.file_id) + change.path

                if change.kind[0] != change.kind[1]:
                    yield ('change kind', change.file_id, change.kind[0], change.kind[1], change.path[0])
                elif change.kind[0] == 'symlink':
                    t_target = self.target_tree.get_symlink_target(change.path[0])
                    w_target = self.work_tree.get_symlink_target(change.path[1])
                    yield ('modify target', change.file_id, change.path[0], t_target,
                           w_target)
                elif change.changed_content:
                    yield ('modify text', change.file_id)

    def shelve_change(self, change):
        """Shelve a change in the iter_shelvable format."""
        if change[0] == 'rename':
            self.shelve_rename(change[1])
        elif change[0] == 'delete file':
            self.shelve_deletion(change[1])
        elif change[0] == 'add file':
            self.shelve_creation(change[1])
        elif change[0] in ('change kind', 'modify text'):
            self.shelve_content_change(change[1])
        elif change[0] == 'modify target':
            self.shelve_modify_target(change[1])
        else:
            raise ValueError('Unknown change kind: "%s"' % change[0])

    def shelve_all(self):
        """Shelve all changes.

        :return: ``True`` if changes were shelved, otherwise ``False``.
        """
        change = None
        for change in self.iter_shelvable():
            self.shelve_change(change)
        return change is not None

    def shelve_rename(self, file_id):
        """Shelve a file rename.

        :param file_id: The file id of the file to shelve the renaming of.
        """
        names, parents = self.renames[file_id]
        w_trans_id = self.work_transform.trans_id_file_id(file_id)
        work_parent = self.work_transform.trans_id_file_id(parents[0])
        self.work_transform.adjust_path(names[0], work_parent, w_trans_id)

        s_trans_id = self.shelf_transform.trans_id_file_id(file_id)
        shelf_parent = self.shelf_transform.trans_id_file_id(parents[1])
        self.shelf_transform.adjust_path(names[1], shelf_parent, s_trans_id)

    def shelve_modify_target(self, file_id):
        """Shelve a change of symlink target.

        :param file_id: The file id of the symlink which changed target.
        :param new_target: The target that the symlink should have due
            to shelving.
        """
        new_path = self.target_tree.id2path(file_id)
        new_target = self.target_tree.get_symlink_target(new_path)
        w_trans_id = self.work_transform.trans_id_file_id(file_id)
        self.work_transform.delete_contents(w_trans_id)
        self.work_transform.create_symlink(new_target, w_trans_id)

        old_path = self.work_tree.id2path(file_id)
        old_target = self.work_tree.get_symlink_target(old_path)
        s_trans_id = self.shelf_transform.trans_id_file_id(file_id)
        self.shelf_transform.delete_contents(s_trans_id)
        self.shelf_transform.create_symlink(old_target, s_trans_id)

    def shelve_lines(self, file_id, new_lines):
        """Shelve text changes to a file, using provided lines.

        :param file_id: The file id of the file to shelve the text of.
        :param new_lines: The lines that the file should have due to shelving.
        """
        w_trans_id = self.work_transform.trans_id_file_id(file_id)
        self.work_transform.delete_contents(w_trans_id)
        self.work_transform.create_file(new_lines, w_trans_id)

        s_trans_id = self.shelf_transform.trans_id_file_id(file_id)
        self.shelf_transform.delete_contents(s_trans_id)
        inverse_lines = self._inverse_lines(new_lines, file_id)
        self.shelf_transform.create_file(inverse_lines, s_trans_id)

    @staticmethod
    def _content_from_tree(tt, tree, file_id):
        trans_id = tt.trans_id_file_id(file_id)
        tt.delete_contents(trans_id)
        transform.create_from_tree(tt, trans_id, tree, tree.id2path(file_id))

    def shelve_content_change(self, file_id):
        """Shelve a kind change or binary file content change.

        :param file_id: The file id of the file to shelve the content change
            of.
        """
        self._content_from_tree(self.work_transform, self.target_tree, file_id)
        self._content_from_tree(self.shelf_transform, self.work_tree, file_id)

    def shelve_creation(self, file_id):
        """Shelve creation of a file.

        This handles content and inventory id.
        :param file_id: The file_id of the file to shelve creation of.
        """
        kind, name, parent, versioned = self.creation[file_id]
        version = not versioned[0]
        self._shelve_creation(self.work_tree, file_id, self.work_transform,
                              self.shelf_transform, kind, name, parent,
                              version)

    def shelve_deletion(self, file_id):
        """Shelve deletion of a file.

        This handles content and inventory id.
        :param file_id: The file_id of the file to shelve deletion of.
        """
        kind, name, parent, versioned = self.deletion[file_id]
        existing_path = self.target_tree.id2path(file_id)
        if not self.work_tree.has_filename(existing_path):
            existing_path = None
        version = not versioned[1]
        self._shelve_creation(self.target_tree, file_id, self.shelf_transform,
                              self.work_transform, kind, name, parent,
                              version, existing_path=existing_path)

    def _shelve_creation(self, tree, file_id, from_transform, to_transform,
                         kind, name, parent, version, existing_path=None):
        w_trans_id = from_transform.trans_id_file_id(file_id)
        if parent is not None and kind is not None:
            from_transform.delete_contents(w_trans_id)
        from_transform.unversion_file(w_trans_id)

        if existing_path is not None:
            s_trans_id = to_transform.trans_id_tree_path(existing_path)
        else:
            s_trans_id = to_transform.trans_id_file_id(file_id)
        if parent is not None:
            s_parent_id = to_transform.trans_id_file_id(parent)
            to_transform.adjust_path(name, s_parent_id, s_trans_id)
            if existing_path is None:
                if kind is None:
                    to_transform.create_file([b''], s_trans_id)
                else:
                    transform.create_from_tree(
                        to_transform, s_trans_id, tree,
                        tree.id2path(file_id))
        if version:
            to_transform.version_file(s_trans_id, file_id=file_id)

    def _inverse_lines(self, new_lines, file_id):
        """Produce a version with only those changes removed from new_lines."""
        target_path = self.target_tree.id2path(file_id)
        target_lines = self.target_tree.get_file_lines(target_path)
        work_path = self.work_tree.id2path(file_id)
        work_lines = self.work_tree.get_file_lines(work_path)
        import patiencediff
        from merge3 import Merge3
        return Merge3(
            new_lines, target_lines, work_lines,
            sequence_matcher=patiencediff.PatienceSequenceMatcher,
            ).merge_lines()

    def finalize(self):
        """Release all resources used by this ShelfCreator."""
        self.work_transform.finalize()
        self.shelf_transform.finalize()

    def transform(self):
        """Shelve changes from working tree."""
        self.work_transform.apply()

    @staticmethod
    def metadata_record(serializer, revision_id, message=None):
        metadata = {b'revision_id': revision_id}
        if message is not None:
            metadata[b'message'] = message.encode('utf-8')
        return serializer.bytes_record(
            bencode.bencode(metadata), ((b'metadata',),))

    def write_shelf(self, shelf_file, message=None):
        """Serialize the shelved changes to a file.

        :param shelf_file: A file-like object to write the shelf to.
        :param message: An optional message describing the shelved changes.
        :return: the filename of the written file.
        """
        transform.resolve_conflicts(self.shelf_transform)
        revision_id = self.target_tree.get_revision_id()
        return self._write_shelf(shelf_file, self.shelf_transform, revision_id,
                                 message)

    @classmethod
    def _write_shelf(cls, shelf_file, transform, revision_id, message=None):
        serializer = pack.ContainerSerialiser()
        shelf_file.write(serializer.begin())
        metadata = cls.metadata_record(serializer, revision_id, message)
        shelf_file.write(metadata)
        for bytes in transform.serialize(serializer):
            shelf_file.write(bytes)
        shelf_file.write(serializer.end())


class Unshelver:
    """Unshelve shelved changes."""

    def __init__(self, tree, base_tree, transform, message):
        """Constructor.

        :param tree: The tree to apply the changes to.
        :param base_tree: The basis to apply the tranform to.
        :param message: A message from the shelved transform.
        """
        self.tree = tree
        self.base_tree = base_tree
        self.transform = transform
        self.message = message

    @staticmethod
    def iter_records(shelf_file):
        parser = pack.ContainerPushParser()
        parser.accept_bytes(shelf_file.read())
        return iter(parser.read_pending_records())

    @staticmethod
    def parse_metadata(records):
        names, metadata_bytes = next(records)
        if names[0] != (b'metadata',):
            raise ShelfCorrupt
        metadata = bencode.bdecode(metadata_bytes)
        message = metadata.get(b'message')
        if message is not None:
            metadata[b'message'] = message.decode('utf-8')
        return metadata

    @classmethod
    def from_tree_and_shelf(klass, tree, shelf_file):
        """Create an Unshelver from a tree and a shelf file.

        :param tree: The tree to apply shelved changes to.
        :param shelf_file: A file-like object containing shelved changes.
        :return: The Unshelver.
        """
        records = klass.iter_records(shelf_file)
        metadata = klass.parse_metadata(records)
        base_revision_id = metadata[b'revision_id']
        try:
            base_tree = tree.revision_tree(base_revision_id)
        except errors.NoSuchRevisionInTree:
            base_tree = tree.branch.repository.revision_tree(base_revision_id)
        tt = base_tree.preview_transform()
        tt.deserialize(records)
        return klass(tree, base_tree, tt, metadata.get(b'message'))

    def make_merger(self):
        """Return a merger that can unshelve the changes."""
        target_tree = self.transform.get_preview_tree()
        merger = merge.Merger.from_uncommitted(self.tree, target_tree,
                                               self.base_tree)
        merger.merge_type = merge.Merge3Merger
        return merger

    def finalize(self):
        """Release all resources held by this Unshelver."""
        self.transform.finalize()


class ShelfManager:
    """Maintain a list of shelved changes."""

    def __init__(self, tree, transport):
        self.tree = tree
        self.transport = transport.clone('shelf')
        self.transport.ensure_base()

    def get_shelf_filename(self, shelf_id):
        return 'shelf-%d' % shelf_id

    def get_shelf_ids(self, filenames):
        matcher = re.compile('shelf-([1-9][0-9]*)')
        shelf_ids = []
        for filename in filenames:
            match = matcher.match(filename)
            if match is not None:
                shelf_ids.append(int(match.group(1)))
        return shelf_ids

    def new_shelf(self):
        """Return a file object and id for a new set of shelved changes."""
        last_shelf = self.last_shelf()
        if last_shelf is None:
            next_shelf = 1
        else:
            next_shelf = last_shelf + 1
        filename = self.get_shelf_filename(next_shelf)
        shelf_file = open(self.transport.local_abspath(filename), 'wb')
        return next_shelf, shelf_file

    def shelve_changes(self, creator, message=None):
        """Store the changes in a ShelfCreator on a shelf."""
        next_shelf, shelf_file = self.new_shelf()
        try:
            creator.write_shelf(shelf_file, message)
        finally:
            shelf_file.close()
        creator.transform()
        return next_shelf

    def read_shelf(self, shelf_id):
        """Return the file associated with a shelf_id for reading.

        :param shelf_id: The id of the shelf to retrive the file for.
        """
        filename = self.get_shelf_filename(shelf_id)
        try:
            return open(self.transport.local_abspath(filename), 'rb')
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            raise NoSuchShelfId(shelf_id)

    def get_unshelver(self, shelf_id):
        """Return an unshelver for a given shelf_id.

        :param shelf_id: The shelf id to return the unshelver for.
        """
        shelf_file = self.read_shelf(shelf_id)
        try:
            return Unshelver.from_tree_and_shelf(self.tree, shelf_file)
        finally:
            shelf_file.close()

    def get_metadata(self, shelf_id):
        """Return the metadata associated with a given shelf_id."""
        shelf_file = self.read_shelf(shelf_id)
        try:
            records = Unshelver.iter_records(shelf_file)
        finally:
            shelf_file.close()
        return Unshelver.parse_metadata(records)

    def delete_shelf(self, shelf_id):
        """Delete the shelved changes for a given id.

        :param shelf_id: id of the shelved changes to delete.
        """
        filename = self.get_shelf_filename(shelf_id)
        self.transport.delete(filename)

    def active_shelves(self):
        """Return a list of shelved changes."""
        active = sorted(self.get_shelf_ids(self.transport.list_dir('.')))
        return active

    def last_shelf(self):
        """Return the id of the last-created shelved change."""
        active = self.active_shelves()
        if len(active) > 0:
            return active[-1]
        else:
            return None



# Copyright (C) 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import doctest
import os
import sys
from io import StringIO

import breezy

from .. import bedding, crash, osutils, plugin, tests
from . import features


class TestApportReporting(tests.TestCaseInTempDir):

    _test_needs_features = [features.apport]

    def test_apport_report(self):
        crash_dir = osutils.joinpath((self.test_base_dir, 'crash'))
        os.mkdir(crash_dir)
        self.overrideEnv('APPORT_CRASH_DIR', crash_dir)
        self.assertEqual(crash_dir, bedding.crash_dir())

        self.overrideAttr(
            breezy.get_global_state(),
            'plugin_warnings',
            {'example': ['Failed to load plugin foo']})

        stderr = StringIO()

        try:
            raise AssertionError("my error")
        except AssertionError as e:
            crash_filename = crash.report_bug_to_apport(sys.exc_info(), stderr)

        # message explaining the crash
        self.assertContainsRe(stderr.getvalue(),
                              "    apport-bug %s" % crash_filename)

        with open(crash_filename) as crash_file:
            report = crash_file.read()

        self.assertContainsRe(report,
                              '(?m)^BrzVersion:')  # should be in the traceback
        self.assertContainsRe(report, 'my error')
        self.assertContainsRe(report, 'AssertionError')
        # see https://bugs.launchpad.net/bzr/+bug/528114
        self.assertContainsRe(report, 'ExecutablePath')
        self.assertContainsRe(report, 'test_apport_report')
        # should also be in there
        self.assertContainsRe(report, '(?m)^CommandLine:')
        self.assertContainsRe(
            report,
            'Failed to load plugin foo')


class TestNonApportReporting(tests.TestCase):
    """Reporting of crash-type bugs without apport.

    This should work in all environments.
    """

    def setup_fake_plugins(self):
        fake = plugin.PlugIn('fake_plugin', plugin)
        fake.version_info = lambda: (1, 2, 3)
        fake_plugins = {"fake_plugin": fake}
        self.overrideAttr(breezy.get_global_state(), 'plugins', fake_plugins)

    def test_report_bug_legacy(self):
        self.setup_fake_plugins()
        err_file = StringIO()
        try:
            raise AssertionError("my error")
        except AssertionError as e:
            crash.report_bug_legacy(sys.exc_info(), err_file)
        report = err_file.getvalue()
        for needle in [
                "brz: ERROR: AssertionError: my error",
                r"Traceback \(most recent call last\):",
                r"plugins: fake_plugin\[1\.2\.3\]",
                ]:
            self.assertContainsRe(report, needle)



# Copyright (C) 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import commands, osutils, tests, trace, ui
from breezy.tests import script


class TestSyntax(tests.TestCase):

    def test_comment_is_ignored(self):
        self.assertEqual([], script._script_to_commands('#comment\n'))

    def test_comment_multiple_lines(self):
        self.assertEqual([
            (['bar'], None, None, None),
            ],
            script._script_to_commands("""
            # this comment is ignored
            # so is this
            # no we run bar
            $ bar
            """))

    def test_trim_blank_lines(self):
        """Blank lines are respected, but trimmed at the start and end.

        Python triple-quoted syntax is going to give stubby/empty blank lines
        right at the start and the end.  These are cut off so that callers don't
        need special syntax to avoid them.

        However we do want to be able to match commands that emit blank lines.
        """
        self.assertEqual([
            (['bar'], None, '\n', None),
            ],
            script._script_to_commands("""
            $bar

            """))

    def test_simple_command(self):
        self.assertEqual([(['cd', 'trunk'], None, None, None)],
                         script._script_to_commands('$ cd trunk'))

    def test_command_with_single_quoted_param(self):
        story = """$ brz commit -m 'two words'"""
        self.assertEqual([(['brz', 'commit', '-m', "'two words'"],
                           None, None, None)],
                         script._script_to_commands(story))

    def test_command_with_double_quoted_param(self):
        story = """$ brz commit -m "two words" """
        self.assertEqual([(['brz', 'commit', '-m', '"two words"'],
                           None, None, None)],
                         script._script_to_commands(story))

    def test_command_with_input(self):
        self.assertEqual(
            [(['cat', '>file'], 'content\n', None, None)],
            script._script_to_commands('$ cat >file\n<content\n'))

    def test_indented(self):
        # scripts are commonly given indented within the test source code, and
        # common indentation is stripped off
        story = """
            $ brz add
            adding file
            adding file2
            """
        self.assertEqual([(['brz', 'add'], None,
                           'adding file\nadding file2\n', None)],
                         script._script_to_commands(story))

    def test_command_with_output(self):
        story = """
$ brz add
adding file
adding file2
"""
        self.assertEqual([(['brz', 'add'], None,
                           'adding file\nadding file2\n', None)],
                         script._script_to_commands(story))

    def test_command_with_error(self):
        story = """
$ brz branch foo
2>brz: ERROR: Not a branch: "foo"
"""
        self.assertEqual([(['brz', 'branch', 'foo'],
                           None, None, 'brz: ERROR: Not a branch: "foo"\n')],
                         script._script_to_commands(story))

    def test_input_without_command(self):
        self.assertRaises(SyntaxError, script._script_to_commands, '<input')

    def test_output_without_command(self):
        self.assertRaises(SyntaxError, script._script_to_commands, '>input')

    def test_command_with_backquotes(self):
        story = """
$ foo = `brz file-id toto`
"""
        self.assertEqual([(['foo', '=', '`brz file-id toto`'],
                           None, None, None)],
                         script._script_to_commands(story))


class TestRedirections(tests.TestCase):

    def _check(self, in_name, out_name, out_mode, remaining, args):
        self.assertEqual(script._scan_redirection_options(args),
                         (in_name, out_name, out_mode, remaining))

    def test_no_redirection(self):
        self._check(None, None, None, [], [])
        self._check(None, None, None, ['foo', 'bar'], ['foo', 'bar'])

    def test_input_redirection(self):
        self._check('foo', None, None, [], ['<foo'])
        self._check('foo', None, None, ['bar'], ['bar', '<foo'])
        self._check('foo', None, None, ['bar'], ['bar', '<', 'foo'])
        self._check('foo', None, None, ['bar'], ['<foo', 'bar'])
        self._check('foo', None, None, ['bar', 'baz'], ['bar', '<foo', 'baz'])

    def test_output_redirection(self):
        self._check(None, 'foo', 'w+', [], ['>foo'])
        self._check(None, 'foo', 'w+', ['bar'], ['bar', '>foo'])
        self._check(None, 'foo', 'w+', ['bar'], ['bar', '>', 'foo'])
        self._check(None, 'foo', 'a+', [], ['>>foo'])
        self._check(None, 'foo', 'a+', ['bar'], ['bar', '>>foo'])
        self._check(None, 'foo', 'a+', ['bar'], ['bar', '>>', 'foo'])

    def test_redirection_syntax_errors(self):
        self._check('', None, None, [], ['<'])
        self._check(None, '', 'w+', [], ['>'])
        self._check(None, '', 'a+', [], ['>>'])
        self._check('>', '', 'a+', [], ['<', '>', '>>'])


class TestExecution(script.TestCaseWithTransportAndScript):

    def test_unknown_command(self):
        """A clear error is reported for commands that aren't recognised

        Testing the attributes of the SyntaxError instance is equivalent to
        using traceback.format_exception_only and comparing with:
          File "<string>", line 1
            foo --frob
            ^
        SyntaxError: Command not found "foo"
        """
        e = self.assertRaises(SyntaxError, self.run_script, "$ foo --frob")
        self.assertContainsRe(e.msg, "not found.*foo")
        self.assertEqual(e.text, "foo --frob")

    def test_blank_output_mismatches_output(self):
        """If you give output, the output must actually be blank.

        See <https://bugs.launchpad.net/bzr/+bug/637830>: previously blank
        output was a wildcard.  Now you must say ... if you want that.
        """
        self.assertRaises(AssertionError,
                          self.run_script,
                          """
            $ echo foo
            """)

    def test_null_output_matches_option(self):
        """If you want null output to be a wild card, you can pass
        null_output_matches_anything to run_script"""
        self.run_script(
            """
            $ echo foo
            """, null_output_matches_anything=True)

    def test_ellipsis_everything(self):
        """A simple ellipsis matches everything."""
        self.run_script("""
        $ echo foo
        ...
        """)

    def test_ellipsis_matches_empty(self):
        self.run_script("""
        $ cd .
        ...
        """)

    def test_stops_on_unexpected_output(self):
        story = """
$ mkdir dir
$ cd dir
The cd command ouputs nothing
"""
        self.assertRaises(AssertionError, self.run_script, story)

    def test_stops_on_unexpected_error(self):
        story = """
$ cat
<Hello
$ brz not-a-command
"""
        self.assertRaises(AssertionError, self.run_script, story)

    def test_continue_on_expected_error(self):
        story = """
$ brz not-a-command
2>..."not-a-command"
"""
        self.run_script(story)

    def test_continue_on_error_output(self):
        # The status matters, not the output
        story = """
$ brz init
...
$ cat >file
<Hello
$ brz add file
...
$ brz commit -m 'adding file'
2>...
"""
        self.run_script(story)

    def test_ellipsis_output(self):
        story = """
$ cat
<first line
<second line
<last line
first line
...
last line
"""
        self.run_script(story)
        story = """
$ brz not-a-command
2>..."not-a-command"
"""
        self.run_script(story)

        story = """
$ brz branch not-a-branch
2>brz: ERROR: Not a branch...not-a-branch/".
"""
        self.run_script(story)


class TestArgumentProcessing(script.TestCaseWithTransportAndScript):

    def test_globing(self):
        self.run_script("""
$ echo cat >cat
$ echo dog >dog
$ cat *
cat
dog
""")

    def test_quoted_globbing(self):
        self.run_script("""
$ echo cat >cat
$ cat '*'
2>*: No such file or directory
""")

    def test_quotes_removal(self):
        self.run_script("""
$ echo 'cat' "dog" '"chicken"' "'dragon'"
cat dog "chicken" 'dragon'
""")

    def test_verbosity_isolated(self):
        """Global verbosity is isolated from commands run in scripts.
        """
        # see also 656694; we should get rid of global verbosity
        self.run_script("""
        $ brz init --quiet a
        """)
        self.assertEqual(trace.is_quiet(), False)


class TestCat(script.TestCaseWithTransportAndScript):

    def test_cat_usage(self):
        self.assertRaises(SyntaxError, self.run_script, 'cat foo <bar')

    def test_cat_input_to_output(self):
        retcode, out, err = self.run_command(['cat'],
                                             'content\n', 'content\n', None)
        self.assertEqual('content\n', out)
        self.assertEqual(None, err)

    def test_cat_file_to_output(self):
        self.build_tree_contents([('file', b'content\n')])
        retcode, out, err = self.run_command(['cat', 'file'],
                                             None, 'content\n', None)
        self.assertEqual('content\n', out)
        self.assertEqual(None, err)

    def test_cat_input_to_file(self):
        retcode, out, err = self.run_command(['cat', '>file'],
                                             'content\n', None, None)
        self.assertFileEqual('content\n', 'file')
        self.assertEqual(None, out)
        self.assertEqual(None, err)
        retcode, out, err = self.run_command(['cat', '>>file'],
                                             'more\n', None, None)
        self.assertFileEqual('content\nmore\n', 'file')
        self.assertEqual(None, out)
        self.assertEqual(None, err)

    def test_cat_file_to_file(self):
        self.build_tree_contents([('file', b'content\n')])
        retcode, out, err = self.run_command(['cat', 'file', '>file2'],
                                             None, None, None)
        self.assertFileEqual(b'content\n', 'file2')

    def test_cat_files_to_file(self):
        self.build_tree_contents([('cat', b'cat\n')])
        self.build_tree_contents([('dog', b'dog\n')])
        retcode, out, err = self.run_command(['cat', 'cat', 'dog', '>file'],
                                             None, None, None)
        self.assertFileEqual(b'cat\ndog\n', 'file')

    def test_cat_bogus_input_file(self):
        self.run_script("""
$ cat <file
2>file: No such file or directory
""")

    def test_cat_bogus_output_file(self):
        self.run_script("""
$ cat >
2>: No such file or directory
""")

    def test_echo_bogus_output_file(self):
        # We need a backing file sysytem for that test so it can't be in
        # TestEcho
        self.run_script("""
$ echo >
2>: No such file or directory
""")


class TestMkdir(script.TestCaseWithTransportAndScript):

    def test_mkdir_usage(self):
        self.assertRaises(SyntaxError, self.run_script, '$ mkdir')
        self.assertRaises(SyntaxError, self.run_script, '$ mkdir foo bar')

    def test_mkdir_jailed(self):
        self.assertRaises(ValueError, self.run_script, '$ mkdir /out-of-jail')
        self.assertRaises(ValueError, self.run_script,
                          '$ mkdir ../out-of-jail')

    def test_mkdir_in_jail(self):
        self.run_script("""
$ mkdir dir
$ cd dir
$ mkdir ../dir2
$ cd ..
""")
        self.assertPathExists('dir')
        self.assertPathExists('dir2')


class TestCd(script.TestCaseWithTransportAndScript):

    def test_cd_usage(self):
        self.assertRaises(SyntaxError, self.run_script, '$ cd foo bar')

    def test_cd_out_of_jail(self):
        self.assertRaises(ValueError, self.run_script, '$ cd /out-of-jail')
        self.assertRaises(ValueError, self.run_script, '$ cd ..')

    def test_cd_dir_and_back_home(self):
        self.assertEqual(self.test_dir, osutils.getcwd())
        self.run_script("""
$ mkdir dir
$ cd dir
""")
        self.assertEqual(osutils.pathjoin(self.test_dir, 'dir'),
                         osutils.getcwd())

        self.run_script('$ cd')
        self.assertEqual(self.test_dir, osutils.getcwd())


class TestBrz(script.TestCaseWithTransportAndScript):

    def test_brz_smoke(self):
        self.run_script("""
            $ brz init branch
            Created a standalone tree (format: ...)
            """)
        self.assertPathExists('branch')


class TestEcho(script.TestCaseWithMemoryTransportAndScript):

    def test_echo_usage(self):
        story = """
$ echo foo
<bar
"""
        self.assertRaises(SyntaxError, self.run_script, story)

    def test_echo_input(self):
        self.assertRaises(SyntaxError, self.run_script, """
            $ echo <foo
            """)

    def test_echo_to_output(self):
        retcode, out, err = self.run_command(['echo'], None, '\n', None)
        self.assertEqual('\n', out)
        self.assertEqual(None, err)

    def test_echo_some_to_output(self):
        retcode, out, err = self.run_command(['echo', 'hello'],
                                             None, 'hello\n', None)
        self.assertEqual('hello\n', out)
        self.assertEqual(None, err)

    def test_echo_more_output(self):
        retcode, out, err = self.run_command(
            ['echo', 'hello', 'happy', 'world'],
            None, 'hello happy world\n', None)
        self.assertEqual('hello happy world\n', out)
        self.assertEqual(None, err)

    def test_echo_appended(self):
        retcode, out, err = self.run_command(['echo', 'hello', '>file'],
                                             None, None, None)
        self.assertEqual(None, out)
        self.assertEqual(None, err)
        self.assertFileEqual(b'hello\n', 'file')
        retcode, out, err = self.run_command(['echo', 'happy', '>>file'],
                                             None, None, None)
        self.assertEqual(None, out)
        self.assertEqual(None, err)
        self.assertFileEqual(b'hello\nhappy\n', 'file')

    def test_empty_line_in_output_is_respected(self):
        self.run_script("""
            $ echo

            $ echo bar
            bar
            """)


class TestRm(script.TestCaseWithTransportAndScript):

    def test_rm_usage(self):
        self.assertRaises(SyntaxError, self.run_script, '$ rm')
        self.assertRaises(SyntaxError, self.run_script, '$ rm -ff foo')

    def test_rm_file(self):
        self.run_script('$ echo content >file')
        self.assertPathExists('file')
        self.run_script('$ rm file')
        self.assertPathDoesNotExist('file')

    def test_rm_file_force(self):
        self.assertPathDoesNotExist('file')
        self.run_script('$ rm -f file')
        self.assertPathDoesNotExist('file')

    def test_rm_files(self):
        self.run_script("""
$ echo content >file
$ echo content >file2
""")
        self.assertPathExists('file2')
        self.run_script('$ rm file file2')
        self.assertPathDoesNotExist('file2')

    def test_rm_dir(self):
        self.run_script('$ mkdir dir')
        self.assertPathExists('dir')
        self.run_script("""
$ rm dir
2>rm: cannot remove 'dir': Is a directory
""")
        self.assertPathExists('dir')

    def test_rm_dir_recursive(self):
        self.run_script("""
$ mkdir dir
$ rm -r dir
""")
        self.assertPathDoesNotExist('dir')


class TestMv(script.TestCaseWithTransportAndScript):

    def test_usage(self):
        self.assertRaises(SyntaxError, self.run_script, '$ mv')
        self.assertRaises(SyntaxError, self.run_script, '$ mv f')
        self.assertRaises(SyntaxError, self.run_script, '$ mv f1 f2 f3')

    def test_move_file(self):
        self.run_script('$ echo content >file')
        self.assertPathExists('file')
        self.run_script('$ mv file new_name')
        self.assertPathDoesNotExist('file')
        self.assertPathExists('new_name')

    def test_move_unknown_file(self):
        self.assertRaises(AssertionError,
                          self.run_script, '$ mv unknown does-not-exist')

    def test_move_dir(self):
        self.run_script("""
$ mkdir dir
$ echo content >dir/file
""")
        self.run_script('$ mv dir new_name')
        self.assertPathDoesNotExist('dir')
        self.assertPathExists('new_name')
        self.assertPathExists('new_name/file')

    def test_move_file_into_dir(self):
        self.run_script("""
$ mkdir dir
$ echo content > file
""")
        self.run_script('$ mv file dir')
        self.assertPathExists('dir')
        self.assertPathDoesNotExist('file')
        self.assertPathExists('dir/file')


class cmd_test_confirm(commands.Command):

    def run(self):
        if ui.ui_factory.get_boolean(
                'Really do it',
                # 'breezy.tests.test_script.confirm',
                # {}
                ):
            self.outf.write('Do it!\n')
        else:
            print('ok, no')


class TestUserInteraction(script.TestCaseWithMemoryTransportAndScript):

    def test_confirm_action(self):
        """You can write tests that demonstrate user confirmation.

        Specifically, ScriptRunner does't care if the output line for the
        prompt isn't terminated by a newline from the program; it's implicitly
        terminated by the input.
        """
        commands.builtin_command_registry.register(cmd_test_confirm)
        self.addCleanup(
            commands.builtin_command_registry.remove, 'test-confirm')
        self.run_script("""
            $ brz test-confirm
            2>Really do it? ([y]es, [n]o): yes
            <y
            Do it!
            $ brz test-confirm
            2>Really do it? ([y]es, [n]o): no
            <n
            ok, no
            """)


class TestShelve(script.TestCaseWithTransportAndScript):

    def setUp(self):
        super().setUp()
        self.run_script("""
            $ brz init test
            Created a standalone tree (format: 2a)
            $ cd test
            $ echo foo > file
            $ brz add
            adding file
            $ brz commit -m 'file added'
            2>Committing to:...test/
            2>added file
            2>Committed revision 1.
            $ echo bar > file
            """)

    def test_shelve(self):
        self.run_script("""
            $ brz shelve -m 'shelve bar'
            2>Shelve? ([y]es, [N]o, [f]inish, [q]uit): yes
            <y
            2>Selected changes:
            2> M  file
            2>Shelve 1 change(s)? ([y]es, [N]o, [f]inish, [q]uit): yes
            <y
            2>Changes shelved with id "1".
            """,
                        null_output_matches_anything=True)
        self.run_script("""
            $ brz shelve --list
              1: shelve bar
            """)

    def test_dont_shelve(self):
        # We intentionally provide no input here to test EOF
        self.run_script((
            "$ brz shelve -m 'shelve bar'\n"
            "2>Shelve? ([y]es, [N]o, [f]inish, [q]uit): \n"
            "2>No changes to shelve.\n"
            ), null_output_matches_anything=True)
        self.run_script("""
            $ brz st
            modified:
              file
            """)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for the _generate_text_key_index API."""


from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository, all_repository_vf_format_scenarios)
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestGenerateTextKeyIndex(TestCaseWithRepository):

    scenarios = all_repository_vf_format_scenarios()

    def test_empty(self):
        repo = self.make_repository('.')
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual({}, repo._generate_text_key_index())



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A collection of extra help information for using brz.

Help topics are meant to be help for items that aren't commands, but will
help brz become fully learnable without referring to a tutorial.

Limited formatting of help text is permitted to make the text useful
both within the reference manual (reStructuredText) and on the screen.
The help text should be reStructuredText with formatting kept to a
minimum and, in particular, no headings. The onscreen renderer applies
the following simple rules before rendering the text:

    1. A '::' appearing on the end of a line is replaced with ':'.
    2. Lines starting with a ':' have it stripped.

These rules mean that literal blocks and field lists respectively can
be used in the help text, producing sensible input to a manual while
rendering on the screen naturally.
"""

import breezy
from breezy import config, i18n, osutils, registry

# Section identifiers (map topics to the right place in the manual)
SECT_COMMAND = "command"
SECT_CONCEPT = "concept"
SECT_HIDDEN = "hidden"
SECT_LIST = "list"
SECT_PLUGIN = "plugin"


class HelpTopicRegistry(registry.Registry):
    """A Registry customized for handling help topics."""

    def register(self, topic, detail, summary, section=SECT_LIST):
        """Register a new help topic.

        :param topic: Name of documentation entry
        :param detail: Function or string object providing detailed
            documentation for topic.  Function interface is detail(topic).
            This should return a text string of the detailed information.
            See the module documentation for details on help text formatting.
        :param summary: String providing single-line documentation for topic.
        :param section: Section in reference manual - see SECT_* identifiers.
        """
        # The detail is stored as the 'object' and the metadata as the info
        info = (summary, section)
        super().register(topic, detail, info=info)

    def register_lazy(self, topic, module_name, member_name, summary,
                      section=SECT_LIST):
        """Register a new help topic, and import the details on demand.

        :param topic: Name of documentation entry
        :param module_name: The module to find the detailed help.
        :param member_name: The member of the module to use for detailed help.
        :param summary: String providing single-line documentation for topic.
        :param section: Section in reference manual - see SECT_* identifiers.
        """
        # The detail is stored as the 'object' and the metadata as the info
        info = (summary, section)
        super().register_lazy(topic, module_name,
                                                     member_name, info=info)

    def get_detail(self, topic):
        """Get the detailed help on a given topic."""
        obj = self.get(topic)
        if callable(obj):
            return obj(topic)
        else:
            return obj

    def get_summary(self, topic):
        """Get the single line summary for the topic."""
        info = self.get_info(topic)
        if info is None:
            return None
        else:
            return info[0]

    def get_section(self, topic):
        """Get the section for the topic."""
        info = self.get_info(topic)
        if info is None:
            return None
        else:
            return info[1]

    def get_topics_for_section(self, section):
        """Get the set of topics in a section."""
        result = set()
        for topic in self.keys():
            if section == self.get_section(topic):
                result.add(topic)
        return result


topic_registry = HelpTopicRegistry()


# ----------------------------------------------------

def _help_on_topics(dummy):
    """Write out the help for topics to outfile"""

    topics = topic_registry.keys()
    lmax = max(len(topic) for topic in topics)

    out = []
    for topic in topics:
        summary = topic_registry.get_summary(topic)
        out.append("%-*s %s\n" % (lmax, topic, summary))
    return ''.join(out)


def _load_from_file(topic_name):
    """Load help from a file.

    Topics are expected to be txt files in breezy.help_topics.
    """
    resource_name = osutils.pathjoin("en", "{}.txt".format(topic_name))
    return osutils.resource_string('breezy.help_topics', resource_name)


def _help_on_revisionspec(name):
    """Generate the help for revision specs."""
    import re

    import breezy.revisionspec

    out = []
    out.append(
        """Revision Identifiers

A revision identifier refers to a specific state of a branch's history.  It
can be expressed in several ways.  It can begin with a keyword to
unambiguously specify a given lookup type; some examples are 'last:1',
'before:yesterday' and 'submit:'.

Alternately, it can be given without a keyword, in which case it will be
checked as a revision number, a tag, a revision id, a date specification, or a
branch specification, in that order.  For example, 'date:today' could be
written as simply 'today', though if you have a tag called 'today' that will
be found first.

If 'REV1' and 'REV2' are revision identifiers, then 'REV1..REV2' denotes a
revision range. Examples: '3647..3649', 'date:yesterday..-1' and
'branch:/path/to/branch1/..branch:/branch2' (note that there are no quotes or
spaces around the '..').

Ranges are interpreted differently by different commands. To the "log" command,
a range is a sequence of log messages, but to the "diff" command, the range
denotes a change between revisions (and not a sequence of changes).  In
addition, "log" considers a closed range whereas "diff" and "merge" consider it
to be open-ended, that is, they include one end but not the other.  For example:
"brz log -r 3647..3649" shows the messages of revisions 3647, 3648 and 3649,
while "brz diff -r 3647..3649" includes the changes done in revisions 3648 and
3649, but not 3647.

The keywords used as revision selection methods are the following:
""")
    details = []
    details.append("\nIn addition, plugins can provide other keywords.")
    details.append(
        "\nA detailed description of each keyword is given below.\n")

    # The help text is indented 4 spaces - this re cleans that up below
    indent_re = re.compile(r'^    ', re.MULTILINE)
    for prefix, i in breezy.revisionspec.revspec_registry.iteritems():
        doc = i.help_txt
        if doc == breezy.revisionspec.RevisionSpec.help_txt:
            summary = "N/A"
            doc = summary + "\n"
        else:
            # Extract out the top line summary from the body and
            # clean-up the unwanted whitespace
            summary, doc = doc.split("\n", 1)
            #doc = indent_re.sub('', doc)
            while (doc[-2:] == '\n\n' or doc[-1:] == ' '):
                doc = doc[:-1]

        # Note: The leading : here are HACKs to get reStructuredText
        # 'field' formatting - we know that the prefix ends in a ':'.
        out.append(":{}\n\t{}".format(i.prefix, summary))
        details.append(":{}\n{}".format(i.prefix, doc))

    return '\n'.join(out + details)


def _help_on_transport(name):
    import textwrap

    from breezy.transport import transport_list_registry

    def add_string(proto, help, maxl, prefix_width=20):
        help_lines = textwrap.wrap(help, maxl - prefix_width,
                                   break_long_words=False)
        line_with_indent = '\n' + ' ' * prefix_width
        help_text = line_with_indent.join(help_lines)
        return "%-20s%s\n" % (proto, help_text)

    def key_func(a):
        return a[:a.rfind("://")]

    protl = []
    decl = []
    protos = transport_list_registry.keys()
    protos.sort(key=key_func)
    for proto in protos:
        shorthelp = transport_list_registry.get_help(proto)
        if not shorthelp:
            continue
        if proto.endswith("://"):
            protl.append(add_string(proto, shorthelp, 79))
        else:
            decl.append(add_string(proto, shorthelp, 79))

    out = "URL Identifiers\n\n" + \
        "Supported URL prefixes::\n\n  " + \
        '  '.join(protl)

    if len(decl):
        out += "\nSupported modifiers::\n\n  " + \
            '  '.join(decl)

    out += """\
\nBreezy supports all of the standard parts within the URL::

  <protocol>://[user[:password]@]host[:port]/[path]

allowing URLs such as::

  http://brzuser:BadPass@brz.example.com:8080/brz/trunk

For brz+ssh:// and sftp:// URLs, Breezy also supports paths that begin
with '~' as meaning that the rest of the path should be interpreted
relative to the remote user's home directory.  For example if the user
``remote`` has a  home directory of ``/home/remote`` on the server
shell.example.com, then::

  brz+ssh://remote@shell.example.com/~/myproject/trunk

would refer to ``/home/remote/myproject/trunk``.

Many commands that accept URLs also accept location aliases too.
See :doc:`location-alias-help` and :doc:`url-special-chars-help`.
"""

    return out


_basic_help = \
    """Breezy %s -- a free distributed version-control tool
https://www.breezy-vcs.org/

Basic commands:
  brz init           makes this directory a versioned branch
  brz branch         make a copy of another branch

  brz add            make files or directories versioned
  brz ignore         ignore a file or pattern
  brz mv             move or rename a versioned file

  brz status         summarize changes in working copy
  brz diff           show detailed diffs

  brz merge          pull in changes from another branch
  brz commit         save some or all changes
  brz send           send changes via email

  brz log            show history of changes
  brz check          validate storage

  brz help init      more help on e.g. init command
  brz help commands  list all commands
  brz help topics    list all help topics
""" % breezy.__version__


_global_options = \
    """Global Options

These options may be used with any command, and may appear in front of any
command.  (e.g. ``brz --profile help``).

--version      Print the version number. Must be supplied before the command.
--no-aliases   Do not process command aliases when running this command.
--builtin      Use the built-in version of a command, not the plugin version.
               This does not suppress other plugin effects.
--no-plugins   Do not process any plugins.
--no-l10n      Do not translate messages.
--concurrency  Number of processes that can be run concurrently (selftest).

--profile      Profile execution using the hotshot profiler.
--lsprof       Profile execution using the lsprof profiler.
--lsprof-file  Profile execution using the lsprof profiler, and write the
               results to a specified file.  If the filename ends with ".txt",
               text format will be used.  If the filename either starts with
               "callgrind.out" or end with ".callgrind", the output will be
               formatted for use with KCacheGrind. Otherwise, the output
               will be a pickle.
--coverage     Generate line coverage report in the specified directory.

-Oname=value   Override the ``name`` config option setting it to ``value`` for
               the duration of the command.  This can be used multiple times if
               several options need to be overridden.

See https://www.breezy-vcs.org/developers/profiling.html for more
information on profiling.

A number of debug flags are also available to assist troubleshooting and
development.  See :doc:`debug-flags-help`.
"""

_standard_options = \
    """Standard Options

Standard options are legal for all commands.

--help, -h     Show help message.
--verbose, -v  Display more information.
--quiet, -q    Only display errors and warnings.

Unlike global options, standard options can be used in aliases.
"""


_checkouts = \
    """Checkouts

Checkouts are source trees that are connected to a branch, so that when
you commit in the source tree, the commit goes into that branch.  They
allow you to use a simpler, more centralized workflow, ignoring some of
Breezy's decentralized features until you want them. Using checkouts
with shared repositories is very similar to working with SVN or CVS, but
doesn't have the same restrictions.  And using checkouts still allows
others working on the project to use whatever workflow they like.

A checkout is created with the brz checkout command (see "help checkout").
You pass it a reference to another branch, and it will create a local copy
for you that still contains a reference to the branch you created the
checkout from (the master branch). Then if you make any commits they will be
made on the other branch first. This creates an instant mirror of your work, or
facilitates lockstep development, where each developer is working together,
continuously integrating the changes of others.

However the checkout is still a first class branch in Breezy terms, so that
you have the full history locally.  As you have a first class branch you can
also commit locally if you want, for instance due to the temporary loss af a
network connection. Use the --local option to commit to do this. All the local
commits will then be made on the master branch the next time you do a non-local
commit.

If you are using a checkout from a shared branch you will periodically want to
pull in all the changes made by others. This is done using the "update"
command. The changes need to be applied before any non-local commit, but
Breezy will tell you if there are any changes and suggest that you use this
command when needed.

It is also possible to create a "lightweight" checkout by passing the
--lightweight flag to checkout. A lightweight checkout is even closer to an
SVN checkout in that it is not a first class branch, it mainly consists of the
working tree. This means that any history operations must query the master
branch, which could be slow if a network connection is involved. Also, as you
don't have a local branch, then you cannot commit locally.

Lightweight checkouts work best when you have fast reliable access to the
master branch. This means that if the master branch is on the same disk or LAN
a lightweight checkout will be faster than a heavyweight one for any commands
that modify the revision history (as only one copy of the branch needs to
be updated). Heavyweight checkouts will generally be faster for any command
that uses the history but does not change it, but if the master branch is on
the same disk then there won't be a noticeable difference.

Another possible use for a checkout is to use it with a treeless repository
containing your branches, where you maintain only one working tree by
switching the master branch that the checkout points to when you want to
work on a different branch.

Obviously to commit on a checkout you need to be able to write to the master
branch. This means that the master branch must be accessible over a writeable
protocol , such as sftp://, and that you have write permissions at the other
end. Checkouts also work on the local file system, so that all that matters is
file permissions.

You can change the master of a checkout by using the "switch" command (see
"help switch").  This will change the location that the commits are sent to.
The "bind" command can also be used to turn a normal branch into a heavy
checkout. If you would like to convert your heavy checkout into a normal
branch so that every commit is local, you can use the "unbind" command. To see
whether or not a branch is bound or not you can use the "info" command. If the
branch is bound it will tell you the location of the bound branch.

Related commands::

  checkout    Create a checkout. Pass --lightweight to get a lightweight
              checkout
  update      Pull any changes in the master branch in to your checkout
  commit      Make a commit that is sent to the master branch. If you have
              a heavy checkout then the --local option will commit to the
              checkout without sending the commit to the master
  switch      Change the master branch that the commits in the checkout will
              be sent to
  bind        Turn a standalone branch into a heavy checkout so that any
              commits will be sent to the master branch
  unbind      Turn a heavy checkout into a standalone branch so that any
              commits are only made locally
  info        Displays whether a branch is bound or unbound. If the branch is
              bound, then it will also display the location of the bound branch
"""

_repositories = \
    """Repositories

Repositories in Breezy are where committed information is stored. There is
a repository associated with every branch.

Repositories are a form of database. Breezy will usually maintain this for
good performance automatically, but in some situations (e.g. when doing
very many commits in a short time period) you may want to ask brz to
optimise the database indices. This can be done by the 'brz pack' command.

By default just running 'brz init' will create a repository within the new
branch but it is possible to create a shared repository which allows multiple
branches to share their information in the same location. When a new branch is
created it will first look to see if there is a containing shared repository it
can use.

When two branches of the same project share a repository, there is
generally a large space saving. For some operations (e.g. branching
within the repository) this translates in to a large time saving.

To create a shared repository use the init-shared-repository command (or the
alias init-shared-repo). This command takes the location of the repository to
create. This means that 'brz init-shared-repository repo' will create a
directory named 'repo', which contains a shared repository. Any new branches
that are created in this directory will then use it for storage.

It is a good idea to create a repository whenever you might create more
than one branch of a project. This is true for both working areas where you
are doing the development, and any server areas that you use for hosting
projects. In the latter case, it is common to want branches without working
trees. Since the files in the branch will not be edited directly there is no
need to use up disk space for a working tree. To create a repository in which
the branches will not have working trees pass the '--no-trees' option to
'init-shared-repository'.

Related commands::

  init-shared-repository   Create a shared repository. Use --no-trees to create
                           one in which new branches won't get a working tree.
"""


_working_trees = \
    """Working Trees

A working tree is the contents of a branch placed on disk so that you can
see the files and edit them. The working tree is where you make changes to a
branch, and when you commit the current state of the working tree is the
snapshot that is recorded in the commit.

When you push a branch to a remote system, a working tree will not be
created. If one is already present the files will not be updated. The
branch information will be updated and the working tree will be marked
as out-of-date. Updating a working tree remotely is difficult, as there
may be uncommitted changes or the update may cause content conflicts that are
difficult to deal with remotely.

If you have a branch with no working tree you can use the 'checkout' command
to create a working tree. If you run 'brz checkout .' from the branch it will
create the working tree. If the branch is updated remotely, you can update the
working tree by running 'brz update' in that directory.

If you have a branch with a working tree that you do not want the 'remove-tree'
command will remove the tree if it is safe. This can be done to avoid the
warning about the remote working tree not being updated when pushing to the
branch. It can also be useful when working with a '--no-trees' repository
(see 'brz help repositories').

If you want to have a working tree on a remote machine that you push to you
can either run 'brz update' in the remote branch after each push, or use some
other method to update the tree during the push. There is an 'rspush' plugin
that will update the working tree using rsync as well as doing a push. There
is also a 'push-and-update' plugin that automates running 'brz update' via SSH
after each push.

Useful commands::

  checkout     Create a working tree when a branch does not have one.
  remove-tree  Removes the working tree from a branch when it is safe to do so.
  update       When a working tree is out of sync with its associated branch
               this will update the tree to match the branch.
"""


_branches = \
    """Branches

A branch consists of the state of a project, including all of its
history. All branches have a repository associated (which is where the
branch history is stored), but multiple branches may share the same
repository (a shared repository). Branches can be copied and merged.

In addition, one branch may be bound to another one.  Binding to another
branch indicates that commits which happen in this branch must also
happen in the other branch.  Breezy ensures consistency by not allowing
commits when the two branches are out of date.  In order for a commit
to succeed, it may be necessary to update the current branch using
``brz update``.

Related commands::

  init    Change a directory into a versioned branch.
  branch  Create a new branch that is a copy of an existing branch.
  merge   Perform a three-way merge.
  bind    Bind a branch to another one.
"""


_standalone_trees = \
    """Standalone Trees

A standalone tree is a working tree with an associated repository. It
is an independently usable branch, with no dependencies on any other.
Creating a standalone tree (via brz init) is the quickest way to put
an existing project under version control.

Related Commands::

  init    Make a directory into a versioned branch.
"""


_status_flags = \
    """Status Flags

Status flags are used to summarise changes to the working tree in a concise
manner.  They are in the form::

   xxx   <filename>

where the columns' meanings are as follows.

Column 1 - versioning/renames::

  + File versioned
  - File unversioned
  R File renamed
  ? File unknown
  X File nonexistent (and unknown to brz)
  C File has conflicts
  P Entry for a pending merge (not a file)

Column 2 - contents::

  N File created
  D File deleted
  K File kind changed
  M File modified

Column 3 - execute::

  * The execute bit was changed
"""


known_env_variables = [
    ("BRZPATH", "Path where brz is to look for shell plugin external commands."),
    ("BRZ_EMAIL", "E-Mail address of the user. Overrides EMAIL."),
    ("EMAIL", "E-Mail address of the user."),
    ("BRZ_EDITOR", "Editor for editing commit messages. Overrides EDITOR."),
    ("EDITOR", "Editor for editing commit messages."),
    ("BRZ_PLUGIN_PATH", "Paths where brz should look for plugins."),
    ("BRZ_DISABLE_PLUGINS", "Plugins that brz should not load."),
    ("BRZ_PLUGINS_AT", "Plugins to load from a directory not in BRZ_PLUGIN_PATH."),
    ("BRZ_HOME", "Directory holding breezy config dir. Overrides HOME."),
    ("BRZ_HOME (Win32)", "Directory holding breezy config dir. Overrides APPDATA and HOME."),
    ("BZR_REMOTE_PATH", "Full name of remote 'brz' command (for brz+ssh:// URLs)."),
    ("BRZ_SSH", "Path to SSH client, or one of paramiko, openssh, sshcorp, plink or lsh."),
    ("BRZ_LOG", "Location of brz.log (use '/dev/null' to suppress log)."),
    ("BRZ_LOG (Win32)", "Location of brz.log (use 'NUL' to suppress log)."),
    ("BRZ_COLUMNS", "Override implicit terminal width."),
    ("BRZ_CONCURRENCY", "Number of processes that can be run concurrently (selftest)"),
    ("BRZ_PROGRESS_BAR", "Override the progress display. Values are 'none' or 'text'."),
    ("BRZ_PDB", "Control whether to launch a debugger on error."),
    ("BRZ_SIGQUIT_PDB",
     "Control whether SIGQUIT behaves normally or invokes a breakin debugger."),
    ("BRZ_TEXTUI_INPUT",
     "Force console input mode for prompts to line-based (instead of char-based)."),
    ]


def _env_variables(topic):
    import textwrap
    ret = ["Environment Variables\n\n"
           "See brz help configuration for more details.\n\n"]
    max_key_len = max([len(k[0]) for k in known_env_variables])
    desc_len = (80 - max_key_len - 2)
    ret.append("=" * max_key_len + " " + "=" * desc_len + "\n")
    for k, desc in known_env_variables:
        ret.append(k + (max_key_len + 1 - len(k)) * " ")
        ret.append("\n".join(textwrap.wrap(
            desc, width=desc_len, subsequent_indent=" " * (max_key_len + 1))))
        ret.append("\n")
    ret += "=" * max_key_len + " " + "=" * desc_len + "\n"
    return "".join(ret)


_files = \
    r"""Files

:On Unix:   ~/.config/breezy/breezy.conf
:On Windows: %APPDATA%\\breezy\\breezy.conf

Contains the user's default configuration. The section ``[DEFAULT]`` is
used to define general configuration that will be applied everywhere.
The section ``[ALIASES]`` can be used to create command aliases for
commonly used options.

A typical config file might look something like::

  [DEFAULT]
  email=John Doe <jdoe@isp.com>

  [ALIASES]
  commit = commit --strict
  log10 = log --short -r -10..-1
"""

_criss_cross = \
    """Criss-Cross

A criss-cross in the branch history can cause the default merge technique
to emit more conflicts than would normally be expected.

In complex merge cases, ``brz merge --lca`` or ``brz merge --weave`` may give
better results.  You may wish to ``brz revert`` the working tree and merge
again.  Alternatively, use ``brz remerge`` on particular conflicted files.

Criss-crosses occur in a branch's history if two branches merge the same thing
and then merge one another, or if two branches merge one another at the same
time.  They can be avoided by having each branch only merge from or into a
designated central branch (a "star topology").

Criss-crosses cause problems because of the way merge works.  Breezy's default
merge is a three-way merger; in order to merge OTHER into THIS, it must
find a basis for comparison, BASE.  Using BASE, it can determine whether
differences between THIS and OTHER are due to one side adding lines, or
from another side removing lines.

Criss-crosses mean there is no good choice for a base.  Selecting the recent
merge points could cause one side's changes to be silently discarded.
Selecting older merge points (which Breezy does) mean that extra conflicts
are emitted.

The ``weave`` merge type is not affected by this problem because it uses
line-origin detection instead of a basis revision to determine the cause of
differences.
"""

_branches_out_of_sync = """Branches Out of Sync

When reconfiguring a checkout, tree or branch into a lightweight checkout,
a local branch must be destroyed.  (For checkouts, this is the local branch
that serves primarily as a cache.)  If the branch-to-be-destroyed does not
have the same last revision as the new reference branch for the lightweight
checkout, data could be lost, so Breezy refuses.

How you deal with this depends on *why* the branches are out of sync.

If you have a checkout and have done local commits, you can get back in sync
by running "brz update" (and possibly "brz commit").

If you have a branch and the remote branch is out-of-date, you can push
the local changes using "brz push".  If the local branch is out of date, you
can do "brz pull".  If both branches have had changes, you can merge, commit
and then push your changes.  If you decide that some of the changes aren't
useful, you can "push --overwrite" or "pull --overwrite" instead.
"""


_storage_formats = \
    """Storage Formats

To ensure that older clients do not access data incorrectly,
Breezy's policy is to introduce a new storage format whenever
new features requiring new metadata are added. New storage
formats may also be introduced to improve performance and
scalability.

The newest format, 2a, is highly recommended. If your
project is not using 2a, then you should suggest to the
project owner to upgrade.


.. note::

   Some of the older formats have two variants:
   a plain one and a rich-root one. The latter include an additional
   field about the root of the tree. There is no performance cost
   for using a rich-root format but you cannot easily merge changes
   from a rich-root format into a plain format. As a consequence,
   moving a project to a rich-root format takes some co-ordination
   in that all contributors need to upgrade their repositories
   around the same time. 2a and all future formats will be
   implicitly rich-root.

See :doc:`current-formats-help` for the complete list of
currently supported formats. See :doc:`other-formats-help` for
descriptions of any available experimental and deprecated formats.
"""


# Register help topics
topic_registry.register("revisionspec", _help_on_revisionspec,
                        "Explain how to use --revision")
topic_registry.register('basic', _basic_help, "Basic commands", SECT_HIDDEN)
topic_registry.register('topics', _help_on_topics, "Topics list", SECT_HIDDEN)


def get_current_formats_topic(topic):
    from breezy import controldir
    return "Current Storage Formats\n\n" + \
        controldir.format_registry.help_topic(topic)


def get_other_formats_topic(topic):
    from breezy import controldir
    return "Other Storage Formats\n\n" + \
        controldir.format_registry.help_topic(topic)


topic_registry.register('current-formats', get_current_formats_topic,
                        'Current storage formats')
topic_registry.register('other-formats', get_other_formats_topic,
                        'Experimental and deprecated storage formats')
topic_registry.register('standard-options', _standard_options,
                        'Options that can be used with any command')
topic_registry.register('global-options', _global_options,
                        'Options that control how Breezy runs')
topic_registry.register('urlspec', _help_on_transport,
                        "Supported transport protocols")
topic_registry.register('status-flags', _status_flags,
                        "Help on status flags")


def get_bugs_topic(topic):
    from breezy import bugtracker
    return ("Bug Tracker Settings\n\n"
            + bugtracker.tracker_registry.help_topic(topic))


topic_registry.register('bugs', get_bugs_topic, 'Bug tracker settings')
topic_registry.register('env-variables', _env_variables,
                        'Environment variable names and values')
topic_registry.register('files', _files,
                        'Information on configuration and log files')
topic_registry.register_lazy('hooks', 'breezy.hooks', 'hooks_help_text',
                             'Points at which custom processing can be added')
topic_registry.register_lazy('location-alias', 'breezy.directory_service',
                             'AliasDirectory.help_text',
                             'Aliases for remembered locations')

# Load some of the help topics from files. Note that topics which reproduce API
# details will tend to skew (quickly usually!) so please seek other solutions
# for such things.
topic_registry.register('authentication', _load_from_file,
                        'Information on configuring authentication')
topic_registry.register('configuration', _load_from_file,
                        'Details on the configuration settings available')
topic_registry.register('conflict-types', _load_from_file,
                        'Types of conflicts and what to do about them')
topic_registry.register('debug-flags', _load_from_file,
                        'Options to show or record debug information')
topic_registry.register('glossary', _load_from_file, 'Glossary')
topic_registry.register('log-formats', _load_from_file,
                        'Details on the logging formats available')
topic_registry.register('missing-extensions', _load_from_file,
                        'What to do when compiled extensions are missing')
topic_registry.register('url-special-chars', _load_from_file,
                        'Special character handling in URLs')


# Register concept topics.
# Note that we might choose to remove these from the online help in the
# future or implement them via loading content from files. In the meantime,
# please keep them concise.
topic_registry.register('branches', _branches,
                        'Information on what a branch is', SECT_CONCEPT)
topic_registry.register('checkouts', _checkouts,
                        'Information on what a checkout is', SECT_CONCEPT)
topic_registry.register('content-filters', _load_from_file,
                        'Conversion of content into/from working trees',
                        SECT_CONCEPT)
topic_registry.register('diverged-branches', _load_from_file,
                        'How to fix diverged branches',
                        SECT_CONCEPT)
topic_registry.register('eol', _load_from_file,
                        'Information on end-of-line handling',
                        SECT_CONCEPT)
topic_registry.register('formats', _storage_formats,
                        'Information on choosing a storage format',
                        SECT_CONCEPT)
topic_registry.register('patterns', _load_from_file,
                        'Information on the pattern syntax',
                        SECT_CONCEPT)
topic_registry.register('repositories', _repositories,
                        'Basic information on shared repositories.',
                        SECT_CONCEPT)
topic_registry.register('rules', _load_from_file,
                        'Information on defining rule-based preferences',
                        SECT_CONCEPT)
topic_registry.register('standalone-trees', _standalone_trees,
                        'Information on what a standalone tree is',
                        SECT_CONCEPT)
topic_registry.register('working-trees', _working_trees,
                        'Information on working trees', SECT_CONCEPT)
topic_registry.register('criss-cross', _criss_cross,
                        'Information on criss-cross merging', SECT_CONCEPT)
topic_registry.register('sync-for-reconfigure', _branches_out_of_sync,
                        'Steps to resolve "out-of-sync" when reconfiguring',
                        SECT_CONCEPT)


class HelpTopicIndex:
    """A index for brz help that returns topics."""

    def __init__(self):
        self.prefix = ''

    def get_topics(self, topic):
        """Search for topic in the HelpTopicRegistry.

        :param topic: A topic to search for. None is treated as 'basic'.
        :return: A list which is either empty or contains a single
            RegisteredTopic entry.
        """
        if topic is None:
            topic = 'basic'
        if topic in topic_registry:
            return [RegisteredTopic(topic)]
        else:
            return []


def _format_see_also(see_also):
    result = ''
    if see_also:
        result += '\n:See also: '
        result += ', '.join(sorted(set(see_also)))
        result += '\n'
    return result


class RegisteredTopic:
    """A help topic which has been registered in the HelpTopicRegistry.

    These topics consist of nothing more than the name of the topic - all
    data is retrieved on demand from the registry.
    """

    def __init__(self, topic):
        """Constructor.

        :param topic: The name of the topic that this represents.
        """
        self.topic = topic

    def get_help_text(self, additional_see_also=None, plain=True):
        """Return a string with the help for this topic.

        :param additional_see_also: Additional help topics to be
            cross-referenced.
        :param plain: if False, raw help (reStructuredText) is
            returned instead of plain text.
        """
        result = topic_registry.get_detail(self.topic)
        result += _format_see_also(additional_see_also)
        if plain:
            result = help_as_plain_text(result)
        i18n.install()
        result = i18n.gettext_per_paragraph(result)
        return result

    def get_help_topic(self):
        """Return the help topic this can be found under."""
        return self.topic


def help_as_plain_text(text):
    """Minimal converter of reStructuredText to plain text."""
    import re

    # Remove the standalone code block marker
    text = re.sub(r"(?m)^\s*::\n\s*$", "", text)
    lines = text.splitlines()
    result = []
    for line in lines:
        if line.startswith(':'):
            line = line[1:]
        elif line.endswith('::'):
            line = line[:-1]
        # Map :doc:`xxx-help` to ``brz help xxx``
        line = re.sub(":doc:`(.+?)-help`", r'``brz help \1``', line)
        result.append(line)
    return "\n".join(result) + "\n"


class ConfigOptionHelpIndex:
    """A help index that returns help topics for config options."""

    def __init__(self):
        self.prefix = 'configuration/'

    def get_topics(self, topic):
        """Search for topic in the registered config options.

        :param topic: A topic to search for.
        :return: A list which is either empty or contains a single
            config.Option entry.
        """
        if topic is None:
            return []
        elif topic.startswith(self.prefix):
            topic = topic[len(self.prefix):]
        if topic in config.option_registry:
            return [config.option_registry.get(topic)]
        else:
            return []



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all Tree's implement get_symlink_target"""

import os

from breezy import osutils, tests
from breezy.tests import features, per_tree


class TestGetSymlinkTarget(per_tree.TestCaseWithTree):

    def get_tree_with_symlinks(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree('tree')
        os.symlink('foo', 'tree/link')
        os.symlink('../bar', 'tree/rel_link')
        os.symlink('/baz/bing', 'tree/abs_link')

        tree.add(['link', 'rel_link', 'abs_link'])
        return self._convert_tree(tree)

    def test_get_symlink_target(self):
        tree = self.get_tree_with_symlinks()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual('foo', tree.get_symlink_target('link'))
        self.assertEqual('../bar',
                         tree.get_symlink_target('rel_link'))
        self.assertEqual('/baz/bing',
                         tree.get_symlink_target('abs_link'))
        self.assertEqual('foo',
                         tree.get_symlink_target('link'))

    def test_get_unicode_symlink_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree('tree')
        target = 'targ\N{Euro Sign}t'
        os.symlink(target, os.fsencode('tree/\u03b2_link'))
        tree.add(['\u03b2_link'])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        actual = tree.get_symlink_target('\u03b2_link')
        self.assertEqual(target, actual)



# Copyright (C) 2007-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for commands related to tags"""

from breezy import branch as _mod_branch
from breezy import errors, lockable_files, lockdir, tag
from breezy.branch import Branch
from breezy.bzr import branch as bzrbranch
from breezy.bzr import bzrdir
from breezy.tests import TestCaseWithTransport, script
from breezy.workingtree import WorkingTree


class TestTagging(TestCaseWithTransport):

    def test_tag_command_help(self):
        out, err = self.run_bzr('help tag')
        self.assertContainsRe(out, 'Create, remove or modify a tag')

    def test_cannot_tag_range(self):
        out, err = self.run_bzr('tag -r1..10 name', retcode=3)
        self.assertContainsRe(err,
                              "Tags can only be placed on a single revision")

    def test_no_tag_name(self):
        out, err = self.run_bzr('tag -d branch', retcode=3)
        self.assertContainsRe(err, 'Please specify a tag name.')

    def test_automatic_tag_name(self):
        def get_tag_name(branch, revid):
            return "mytag"
        Branch.hooks.install_named_hook('automatic_tag_name',
                                        get_tag_name, 'get tag name')
        out, err = self.run_bzr('tag -d branch')
        self.assertContainsRe(err, 'Created tag mytag.')

    def test_tag_current_rev(self):
        t = self.make_branch_and_tree('branch')
        t.commit(allow_pointless=True, message='initial commit',
                 rev_id=b'first-revid')
        # make a tag through the command line
        out, err = self.run_bzr('tag -d branch NEWTAG')
        self.assertContainsRe(err, 'Created tag NEWTAG.')
        # tag should be observable through the api
        self.assertEqual(t.branch.tags.get_tag_dict(),
                         dict(NEWTAG=b'first-revid'))
        # can also create tags using -r
        self.run_bzr('tag -d branch tag2 -r1')
        self.assertEqual(t.branch.tags.lookup_tag('tag2'), b'first-revid')
        # regression test: make sure a unicode revision from the user
        # gets turned into a str object properly. The use of a unicode
        # object for the revid is intentional.
        self.run_bzr(['tag', '-d', 'branch', 'tag3', '-rrevid:first-revid'])
        self.assertEqual(t.branch.tags.lookup_tag('tag3'), b'first-revid')
        # can also delete an existing tag
        out, err = self.run_bzr('tag --delete -d branch tag2')
        # cannot replace an existing tag normally
        out, err = self.run_bzr('tag -d branch NEWTAG -r0', retcode=3)
        self.assertContainsRe(err, 'Tag NEWTAG already exists\\.')
        # ... but can if you use --force
        out, err = self.run_bzr('tag -d branch NEWTAG --force -r0')
        self.assertEqual("Updated tag NEWTAG.\n", err)

    def test_tag_same_revision(self):
        t = self.make_branch_and_tree('branch')
        t.commit(allow_pointless=True, message='initial commit',
                 rev_id=b'first-revid')
        t.commit(allow_pointless=True, message='second commit',
                 rev_id=b'second-revid')
        out, err = self.run_bzr('tag -rrevid:first-revid -d branch NEWTAG')
        out, err = self.run_bzr('tag -rrevid:first-revid -d branch NEWTAG')
        self.assertContainsRe(err,
                              'Tag NEWTAG already exists for that revision\\.')
        out, err = self.run_bzr('tag -rrevid:second-revid -d branch NEWTAG',
                                retcode=3)
        self.assertContainsRe(err, 'Tag NEWTAG already exists\\.')

    def test_tag_delete_requires_name(self):
        out, err = self.run_bzr('tag -d branch', retcode=3)
        self.assertContainsRe(err, 'Please specify a tag name\\.')

    def test_branch_push_pull_merge_copies_tags(self):
        t = self.make_branch_and_tree('branch1')
        t.commit(allow_pointless=True, message='initial commit',
                 rev_id=b'first-revid')
        b1 = t.branch
        b1.tags.set_tag('tag1', b'first-revid')
        # branching copies the tag across
        self.run_bzr('branch branch1 branch2')
        b2 = Branch.open('branch2')
        self.assertEqual(b2.tags.lookup_tag('tag1'), b'first-revid')
        # make a new tag and pull it
        b1.tags.set_tag('tag2', b'twa')
        self.run_bzr('pull -d branch2 branch1')
        self.assertEqual(b2.tags.lookup_tag('tag2'), b'twa')
        # make a new tag and push it
        b1.tags.set_tag('tag3', b'san')
        self.run_bzr('push -d branch1 branch2')
        self.assertEqual(b2.tags.lookup_tag('tag3'), b'san')
        # make a new tag and merge it
        t.commit(allow_pointless=True, message='second commit',
                 rev_id=b'second-revid')
        t2 = WorkingTree.open('branch2')
        t2.commit(allow_pointless=True, message='commit in second')
        b1.tags.set_tag('tag4', b'second-revid')
        self.run_bzr('merge -d branch2 branch1')
        self.assertEqual(b2.tags.lookup_tag('tag4'), b'second-revid')
        # pushing to a new location copies the tag across
        self.run_bzr('push -d branch1 branch3')
        b3 = Branch.open('branch3')
        self.assertEqual(b3.tags.lookup_tag('tag1'), b'first-revid')

    def make_master_and_checkout(self):
        builder = self.make_branch_builder('master')
        builder.build_commit(message='Initial commit.', rev_id=b'rev-1')
        master = builder.get_branch()
        child = master.create_checkout(self.get_url('child'))
        return master, child

    def make_fork(self, branch):
        fork = branch.create_clone_on_transport(self.get_transport('fork'))
        self.addCleanup(fork.lock_write().unlock)
        with fork.basis_tree().preview_transform() as tt:
            tt.commit(fork, message='Commit in fork.', revision_id=b'fork-0')
        with fork.basis_tree().preview_transform() as tt:
            tt.commit(fork, message='Commit in fork.', revision_id=b'fork-1')
        return fork

    def test_merge_without_commit_does_not_propagate_tags_to_master(self):
        """'brz merge' alone does not propagate tags to a master branch.

        (If the user runs 'brz commit', then that is when the tags from the
        merge are propagated.)
        """
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag('new-tag', fork.last_revision())
        self.run_bzr(['merge', '../fork'], working_dir='child')
        self.assertEqual({}, master.tags.get_tag_dict())

    def test_commit_in_heavyweight_checkout_copies_tags_to_master(self):
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag('new-tag', fork.last_revision())
        fork.tags.set_tag('non-ancestry-tag', b'fork-0')
        fork.tags.set_tag('absent-tag', b'absent-rev')
        script.run_script(self, """
            $ cd child
            $ brz merge ../fork
            $ brz commit -m "Merge fork."
            2>Committing to: .../master/
            2>Committed revision 2.
            """, null_output_matches_anything=True)
        # Merge copied the tag to child and commit propagated it to master
        expected_tag_dict = {
            'new-tag': fork.last_revision(),
            'non-ancestry-tag': b'fork-0',
            'absent-tag': b'absent-rev',
            }
        self.assertEqual(expected_tag_dict, child.branch.tags.get_tag_dict())
        self.assertEqual(expected_tag_dict, master.tags.get_tag_dict())
        # Revisions not in ancestry but named in tags are present
        child.branch.repository.get_revision(b'fork-0')
        master.repository.get_revision(b'fork-0')

    def test_commit_in_heavyweight_checkout_reports_tag_conflict(self):
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag('new-tag', fork.last_revision())
        master_r1 = master.last_revision()
        master.tags.set_tag('new-tag', master_r1)
        script.run_script(self, """
            $ cd child
            $ brz merge ../fork
            $ brz commit -m "Merge fork."
            2>Committing to: .../master/
            2>Conflicting tags in bound branch:
            2>    new-tag
            2>Committed revision 2.
            """, null_output_matches_anything=True)
        # Merge copied the tag to child.  master's conflicting tag is unchanged.
        self.assertEqual(
            {'new-tag': fork.last_revision()}, child.branch.tags.get_tag_dict())
        self.assertEqual(
            {'new-tag': master_r1}, master.tags.get_tag_dict())

    def test_list_tags(self):
        tree1 = self.make_branch_and_tree('branch1')
        tree1.commit(allow_pointless=True, message='revision 1',
                     rev_id=b'revid-1', timestamp=10)
        tree1.commit(allow_pointless=True, message='revision 2',
                     rev_id=b'revid-2', timestamp=15)

        b1 = tree1.branch
        # note how the tag for revid-1 sorts after the one for revid-2
        b1.tags.set_tag('tag1\u30d0', b'revid-2')
        # not present in repository
        b1.tags.set_tag('tag10\u30d0', b'missing')
        b1.tags.set_tag('tag2\u30d0', b'revid-1')

        # natural order
        out, err = self.run_bzr_raw('tags -d branch1', encoding='utf-8')
        self.assertEqual(err, b'')
        self.assertContainsRe(out, ('^tag1\u30d0  *2\ntag2\u30d0  *1\n' +
                                    'tag10\u30d0 *\\?\n').encode('utf-8'))

        # lexicographical order
        out, err = self.run_bzr_raw('tags --sort=alpha -d branch1',
                                    encoding='utf-8')
        self.assertEqual(err, b'')
        self.assertContainsRe(out, ('^tag10\u30d0  *\\?\ntag1\u30d0  *2\n' +
                                    'tag2\u30d0 *1\n').encode('utf-8'))

        out, err = self.run_bzr_raw('tags --sort=alpha --show-ids -d branch1',
                                    encoding='utf-8')
        self.assertEqual(err, b'')
        self.assertContainsRe(out, ('^tag10\u30d0  *missing\n' +
                                    'tag1\u30d0  *revid-2\ntag2\u30d0 *revid-1\n').encode('utf-8'))

        # chronological order
        out, err = self.run_bzr_raw('tags --sort=time -d branch1',
                                    encoding='utf-8')
        self.assertEqual(err, b'')
        self.assertContainsRe(out, ('^tag2\u30d0  *1\ntag1\u30d0  *2\n' +
                                    'tag10\u30d0 *\\?\n').encode('utf-8'))

        out, err = self.run_bzr_raw('tags --sort=time --show-ids -d branch1',
                                    encoding='utf-8')
        self.assertEqual(err, b'')
        self.assertContainsRe(out, ('^tag2\u30d0  *revid-1\n' +
                                    'tag1\u30d0  *revid-2\ntag10\u30d0 *missing\n').encode('utf-8'))

        # now test dotted revnos
        tree2 = tree1.controldir.sprout('branch2').open_workingtree()
        tree1.commit(allow_pointless=True, message='revision 3 in branch1',
                     rev_id=b'revid-3a')
        tree2.commit(allow_pointless=True, message='revision 3 in branch2',
                     rev_id=b'revid-3b')

        b2 = tree2.branch
        b2.tags.set_tag('tagD', b'revid-3b')
        self.run_bzr('merge -d branch1 branch2')
        tree1.commit('merge', rev_id=b'revid-4')

        out, err = self.run_bzr('tags -d branch1', encoding='utf-8')
        self.assertEqual(err, '')
        self.assertContainsRe(out, r'tagD  *2\.1\.1\n')
        out, err = self.run_bzr('tags -d branch2', encoding='utf-8')
        self.assertEqual(err, '')
        self.assertContainsRe(out, r'tagD  *3\n')

    def test_list_tags_dotted_revnos_unsupported(self):
        class TrimmedBranch(bzrbranch.BzrBranch6):

            def revision_id_to_dotted_revno(self, revid):
                raise errors.UnsupportedOperation(
                    self.revision_id_to_dotted_revno, self)

        class TrimmedBranchFormat(bzrbranch.BzrBranchFormat6):

            def _branch_class(self):
                return TrimmedBranch

            @classmethod
            def get_format_string(cls):
                return b"Trimmed Branch"


        _mod_branch.format_registry.register(TrimmedBranchFormat())
        self.addCleanup(_mod_branch.format_registry.remove,
                        TrimmedBranchFormat())

        dir_format = bzrdir.BzrDirMetaFormat1()
        dir_format._branch_format = TrimmedBranchFormat()

        tree = self.make_branch_and_tree('branch', format=dir_format)
        self.assertFileEqual('Trimmed Branch', 'branch/.bzr/branch/format')
        rev1 = tree.commit("rev1")
        tree.branch.tags.set_tag("mytag", rev1)
        out, err = self.run_bzr('tags -d branch', encoding='utf-8')
        self.assertEqual(out, 'mytag                ?\n')

    def test_list_tags_revision_filtering(self):
        tree1 = self.make_branch_and_tree('.')
        tree1.commit(allow_pointless=True, message='revision 1',
                     rev_id=b'revid-1')
        tree1.commit(allow_pointless=True, message='revision 2',
                     rev_id=b'revid-2')
        tree1.commit(allow_pointless=True, message='revision 3',
                     rev_id=b'revid-3')
        tree1.commit(allow_pointless=True, message='revision 4',
                     rev_id=b'revid-4')
        b1 = tree1.branch
        b1.tags.set_tag('tag 1', b'revid-1')
        b1.tags.set_tag('tag 2', b'revid-2')
        b1.tags.set_tag('tag 3', b'revid-3')
        b1.tags.set_tag('tag 4', b'revid-4')
        self._check_tag_filter('', (1, 2, 3, 4))
        self._check_tag_filter('-r ..', (1, 2, 3, 4))
        self._check_tag_filter('-r ..2', (1, 2))
        self._check_tag_filter('-r 2..', (2, 3, 4))
        self._check_tag_filter('-r 2..3', (2, 3))
        self._check_tag_filter('-r 3..2', ())
        self.run_bzr_error(args="tags -r 123",
                           error_regexes=["brz: ERROR: Requested revision: '123' "
                                          "does not exist in branch:"])
        self.run_bzr_error(args="tags -r ..123",
                           error_regexes=["brz: ERROR: Requested revision: '123' "
                                          "does not exist in branch:"])
        self.run_bzr_error(args="tags -r 123.123",
                           error_regexes=["brz: ERROR: Requested revision: '123.123' "
                                          "does not exist in branch:"])

    def test_sort_tags_custom(self):
        def sort_by_dots(branch, tags):
            def sort_key(tag_and_revid):
                return tag_and_revid[0].count(".")
            tags.sort(key=sort_key)

        # Register a custom sort method
        tag.tag_sort_methods.register("dots", sort_by_dots, "Sort by dots.")
        self.addCleanup(tag.tag_sort_methods.remove, "dots")

        tree1 = self.make_branch_and_tree('branch1')
        tree1.commit(allow_pointless=True, message='revision 1',
                     rev_id=b'revid-1', timestamp=10)
        tree1.commit(allow_pointless=True, message='revision 2',
                     rev_id=b'revid-2', timestamp=15)

        b1 = tree1.branch

        b1.tags.set_tag('tag..', b'revid-2')
        b1.tags.set_tag('tag....', b'missing')  # not present in repository
        b1.tags.set_tag('tag.', b'revid-1')
        b1.tags.set_tag('tag...', b'revid-1')
        b1.tags.set_tag('tag....', b'revid-1')

        # sorted by number of dots
        out, err = self.run_bzr('tags --sort=dots -d branch1')
        self.assertEqual(err, '')
        self.assertEqual([
            'tag.                 1',
            'tag..                2',
            'tag...               1',
            'tag....              1'],
            out.splitlines())

    def _check_tag_filter(self, argstr, expected_revnos):
        # upper bound of laziness
        out, err = self.run_bzr('tags ' + argstr)
        self.assertEqual(err, '')
        self.assertContainsRe(out, "^" + ''.join(["tag {} +{}\n".format(
            revno, revno) for revno in expected_revnos]) + "$")

    def test_conflicting_tags(self):
        # setup two empty branches with different tags
        t1 = self.make_branch_and_tree('one')
        t2 = self.make_branch_and_tree('two')
        b1 = t1.branch
        b2 = t2.branch
        tagname = '\u30d0zaar'
        b1.tags.set_tag(tagname, b'revid1')
        b2.tags.set_tag(tagname, b'revid2')
        # push should give a warning about the tags
        out, err = self.run_bzr('push -d one two', encoding='utf-8')
        self.assertContainsRe(out, 'Conflicting tags:\n.*' + tagname)
        # pull should give a warning about the tags
        out, err = self.run_bzr('pull -d one two', encoding='utf-8',
                                retcode=1)
        self.assertContainsRe(out,
                              'Conflicting tags:\n.*' + tagname)
        # merge should give a warning about the tags -- not implemented yet
        ## out, err = self.run_bzr('merge -d one two', encoding='utf-8')
        # self.assertContainsRe(out,
        # 'Conflicting tags:\n.*' + tagname.encode('utf-8'))

    def test_tag_quiet(self):
        t1 = self.make_branch_and_tree('')
        out, err = self.run_bzr('tag --quiet test1')
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_tag_delete_quiet(self):
        t1 = self.make_branch_and_tree('')
        self.run_bzr('tag test1')
        out, err = self.run_bzr('tag --delete --quiet test1')
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_tags_with_mainline_ghosts(self):
        tree = self.make_branch_and_tree('tree1')
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        tree.add('')
        tree.commit('msg1', rev_id=b'rev1')
        tree.commit('msg2', rev_id=b'rev2')
        tree.branch.tags.set_tag('unknown', b'out-of-mainline')
        tree.branch.tags.set_tag('ghost', b'spooky')
        tree.branch.tags.set_tag('tag1', b'rev1')
        tree.branch.tags.set_tag('tag2', b'rev2')

        out, err = self.run_bzr('tags -d tree1', encoding='utf-8')
        self.assertEqual(out,
                         'ghost                ?\n'
                         'tag1                 1\n'
                         'tag2                 2\n'
                         'unknown              ?\n')
        self.assertEqual('', err)

    def test_tag_unsupported(self):
        tree = self.make_branch_and_tree('tree', format='dirstate')
        out, err = self.run_bzr('tag -d tree blah', retcode=3)
        self.assertEqual(out, '')
        self.assertContainsRe(
            err,
            'brz: ERROR: Tags not supported by BzrBranch5\\(.*\\/tree\\/\\); '
            'you may be able to use \'brz upgrade file:\\/\\/.*\\/tree\\/\'.')



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from io import StringIO

from breezy import osutils, trace

from .bzr.inventorytree import InventoryTreeChange


class TreeDelta:
    """Describes changes from one tree to another.

    Contains seven lists with TreeChange objects.

    added
    removed
    renamed
    copied
    kind_changed
    modified
    unchanged
    unversioned

    Each id is listed only once.

    Files that are both modified and renamed or copied are listed only in
    renamed or copied, with the text_modified flag true. The text_modified
    applies either to the content of the file or the target of the
    symbolic link, depending of the kind of file.

    Files are only considered renamed if their name has changed or
    their parent directory has changed.  Renaming a directory
    does not count as renaming all its contents.

    The lists are normally sorted when the delta is created.
    """

    def __init__(self):
        self.added = []
        self.removed = []
        self.renamed = []
        self.copied = []
        self.kind_changed = []
        self.modified = []
        self.unchanged = []
        self.unversioned = []
        self.missing = []

    def __eq__(self, other):
        if not isinstance(other, TreeDelta):
            return False
        return self.added == other.added \
            and self.removed == other.removed \
            and self.renamed == other.renamed \
            and self.copied == other.copied \
            and self.modified == other.modified \
            and self.unchanged == other.unchanged \
            and self.kind_changed == other.kind_changed \
            and self.unversioned == other.unversioned

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "TreeDelta(added=%r, removed=%r, renamed=%r," \
            " copied=%r, kind_changed=%r, modified=%r, unchanged=%r," \
            " unversioned=%r)" % (
                self.added, self.removed, self.renamed, self.copied,
                self.kind_changed, self.modified, self.unchanged,
                self.unversioned)

    def has_changed(self):
        return bool(self.modified
                    or self.added
                    or self.removed
                    or self.renamed
                    or self.copied
                    or self.kind_changed)

    def get_changes_as_text(self, show_ids=False, show_unchanged=False,
                            short_status=False):
        output = StringIO()
        report_delta(output, self, short_status, show_ids, show_unchanged)
        return output.getvalue()


def _compare_trees(old_tree, new_tree, want_unchanged, specific_files,
                   include_root, extra_trees=None,
                   require_versioned=False, want_unversioned=False):
    """Worker function that implements Tree.changes_from."""
    delta = TreeDelta()
    # mutter('start compare_trees')

    for change in new_tree.iter_changes(
            old_tree, want_unchanged, specific_files, extra_trees=extra_trees,
            require_versioned=require_versioned,
            want_unversioned=want_unversioned):
        if change.versioned == (False, False):
            delta.unversioned.append(change)
            continue
        if not include_root and (None, None) == change.parent_id:
            continue
        fully_present = tuple(
            (change.versioned[x] and change.kind[x] is not None)
            for x in range(2))
        if fully_present[0] != fully_present[1]:
            if fully_present[1] is True:
                delta.added.append(change)
            else:
                if change.kind[0] == 'symlink' and not new_tree.supports_symlinks():
                    trace.warning(
                        'Ignoring "%s" as symlinks '
                        'are not supported on this filesystem.' % (change.path[0],))
                else:
                    delta.removed.append(change)
        elif fully_present[0] is False:
            delta.missing.append(change)
        elif change.name[0] != change.name[1] or change.parent_id[0] != change.parent_id[1]:
            # If the name changes, or the parent_id changes, we have a rename or copy
            # (if we move a parent, that doesn't count as a rename for the
            # file)
            if change.copied:
                delta.copied.append(change)
            else:
                delta.renamed.append(change)
        elif change.kind[0] != change.kind[1]:
            delta.kind_changed.append(change)
        elif change.changed_content or change.executable[0] != change.executable[1]:
            delta.modified.append(change)
        else:
            delta.unchanged.append(change)

    def change_key(change):
        if change.path[0] is None:
            path = change.path[1]
        else:
            path = change.path[0]
        return (path, change.file_id)

    delta.removed.sort(key=change_key)
    delta.added.sort(key=change_key)
    delta.renamed.sort(key=change_key)
    delta.copied.sort(key=change_key)
    delta.missing.sort(key=change_key)
    # TODO: jam 20060529 These lists shouldn't need to be sorted
    #       since we added them in alphabetical order.
    delta.modified.sort(key=change_key)
    delta.unchanged.sort(key=change_key)
    delta.unversioned.sort(key=change_key)

    return delta


class _ChangeReporter:
    """Report changes between two trees"""

    def __init__(self, output=None, suppress_root_add=True,
                 output_file=None, unversioned_filter=None, view_info=None,
                 classify=True):
        """Constructor

        :param output: a function with the signature of trace.note, i.e.
            accepts a format and parameters.
        :param supress_root_add: If true, adding the root will be ignored
            (i.e. when a tree has just been initted)
        :param output_file: If supplied, a file-like object to write to.
            Only one of output and output_file may be supplied.
        :param unversioned_filter: A filter function to be called on
            unversioned files. This should return True to ignore a path.
            By default, no filtering takes place.
        :param view_info: A tuple of view_name,view_files if only
            items inside a view are to be reported on, or None for
            no view filtering.
        :param classify: Add special symbols to indicate file kind.
        """
        if output_file is not None:
            if output is not None:
                raise BzrError('Cannot specify both output and output_file')

            def output(fmt, *args):
                output_file.write((fmt % args) + '\n')
        self.output = output
        if self.output is None:
            from . import trace
            self.output = trace.note
        self.suppress_root_add = suppress_root_add
        self.modified_map = {'kind changed': 'K',
                             'unchanged': ' ',
                             'created': 'N',
                             'modified': 'M',
                             'deleted': 'D',
                             'missing': '!',
                             }
        self.versioned_map = {'added': '+',  # versioned target
                              'unchanged': ' ',  # versioned in both
                              'removed': '-',  # versioned in source
                              'unversioned': '?',  # versioned in neither
                              }
        self.unversioned_filter = unversioned_filter
        if classify:
            self.kind_marker = osutils.kind_marker
        else:
            self.kind_marker = lambda kind: ''
        if view_info is None:
            self.view_name = None
            self.view_files = []
        else:
            self.view_name = view_info[0]
            self.view_files = view_info[1]
            self.output("Operating on whole tree but only reporting on "
                        "'%s' view." % (self.view_name,))

    def report(self, paths, versioned, renamed, copied, modified, exe_change,
               kind):
        """Report one change to a file

        :param path: The old and new paths as generated by Tree.iter_changes.
        :param versioned: may be 'added', 'removed', 'unchanged', or
            'unversioned.
        :param renamed: may be True or False
        :param copied: may be True or False
        :param modified: may be 'created', 'deleted', 'kind changed',
            'modified' or 'unchanged'.
        :param exe_change: True if the execute bit has changed
        :param kind: A pair of file kinds, as generated by Tree.iter_changes.
            None indicates no file present.
        """
        if trace.is_quiet():
            return
        if paths[1] == '' and versioned == 'added' and self.suppress_root_add:
            return
        if self.view_files and not osutils.is_inside_any(self.view_files,
                                                         paths[1]):
            return
        if versioned == 'unversioned':
            # skip ignored unversioned files if needed.
            if self.unversioned_filter is not None:
                if self.unversioned_filter(paths[1]):
                    return
            # dont show a content change in the output.
            modified = 'unchanged'
        # we show both paths in the following situations:
        # the file versioning is unchanged AND
        # ( the path is different OR
        #   the kind is different)
        if (versioned == 'unchanged' and
                (renamed or copied or modified == 'kind changed')):
            if renamed or copied:
                # on a rename or copy, we show old and new
                old_path, path = paths
            else:
                # if it's not renamed or copied, we're showing both for kind
                # changes so only show the new path
                old_path, path = paths[1], paths[1]
            # if the file is not missing in the source, we show its kind
            # when we show two paths.
            if kind[0] is not None:
                old_path += self.kind_marker(kind[0])
            old_path += " => "
        elif versioned == 'removed':
            # not present in target
            old_path = ""
            path = paths[0]
        else:
            old_path = ""
            path = paths[1]
        if renamed:
            rename = "R"
        elif copied:
            rename = "C"
        else:
            rename = self.versioned_map[versioned]
        # we show the old kind on the new path when the content is deleted.
        if modified == 'deleted':
            path += self.kind_marker(kind[0])
        # otherwise we always show the current kind when there is one
        elif kind[1] is not None:
            path += self.kind_marker(kind[1])
        if exe_change:
            exe = '*'
        else:
            exe = ' '
        self.output("%s%s%s %s%s", rename, self.modified_map[modified], exe,
                    old_path, path)


def report_changes(change_iterator, reporter):
    """Report the changes from a change iterator.

    This is essentially a translation from low-level to medium-level changes.
    Further processing may be required to produce a human-readable output.
    Unfortunately, some tree-changing operations are very complex
    :change_iterator: an iterator or sequence of changes in the format
        generated by Tree.iter_changes
    :param reporter: The _ChangeReporter that will report the changes.
    """
    versioned_change_map = {
        (True, True): 'unchanged',
        (True, False): 'removed',
        (False, True): 'added',
        (False, False): 'unversioned',
        }

    def path_key(change):
        if change.path[0] is not None:
            path = change.path[0]
        else:
            path = change.path[1]
        return osutils.splitpath(path)
    for change in sorted(change_iterator, key=path_key):
        exe_change = False
        # files are "renamed" if they are moved or if name changes, as long
        # as it had a value
        if change.copied:
            copied = True
            renamed = False
        elif change.renamed:
            renamed = True
            copied = False
        else:
            copied = False
            renamed = False
        if change.kind[0] != change.kind[1]:
            if change.kind[0] is None:
                modified = "created"
            elif change.kind[1] is None:
                modified = "deleted"
            else:
                modified = "kind changed"
        else:
            if change.changed_content:
                modified = "modified"
            elif change.kind[0] is None:
                modified = "missing"
            else:
                modified = "unchanged"
            if change.kind[1] == "file":
                exe_change = (change.executable[0] != change.executable[1])
        versioned_change = versioned_change_map[change.versioned]
        reporter.report(change.path, versioned_change, renamed, copied, modified,
                        exe_change, change.kind)


def report_delta(to_file, delta, short_status=False, show_ids=False,
                 show_unchanged=False, indent='', predicate=None, classify=True):
    """Output this delta in status-like form to to_file.

    :param to_file: A file-like object where the output is displayed.

    :param delta: A TreeDelta containing the changes to be displayed

    :param short_status: Single-line status if True.

    :param show_ids: Output the file ids if True.

    :param show_unchanged: Output the unchanged files if True.

    :param indent: Added at the beginning of all output lines (for merged
        revisions).

    :param predicate: A callable receiving a path returning True if the path
        should be displayed.

    :param classify: Add special symbols to indicate file kind.
    """

    def decorate_path(path, kind, meta_modified=None):
        if not classify:
            return path
        if kind == 'directory':
            path += '/'
        elif kind == 'symlink':
            path += '@'
        if meta_modified:
            path += '*'
        return path

    def show_more_renamed(item):
        dec_new_path = decorate_path(item.path[1], item.kind[1], item.meta_modified())
        to_file.write(' => %s' % dec_new_path)
        if item.changed_content or item.meta_modified():
            extra_modified.append(InventoryTreeChange(
                item.file_id, (item.path[1], item.path[1]),
                item.changed_content,
                item.versioned,
                (item.parent_id[1], item.parent_id[1]),
                (item.name[1], item.name[1]),
                (item.kind[1], item.kind[1]),
                item.executable))

    def show_more_kind_changed(item):
        to_file.write(' ({} => {})'.format(item.kind[0], item.kind[1]))

    def show_path(path, kind, meta_modified,
                  default_format, with_file_id_format):
        dec_path = decorate_path(path, kind, meta_modified)
        if show_ids:
            to_file.write(with_file_id_format % dec_path)
        else:
            to_file.write(default_format % dec_path)

    def show_list(files, long_status_name, short_status_letter,
                  default_format='%s', with_file_id_format='%-30s',
                  show_more=None):
        if files:
            header_shown = False
            if short_status:
                prefix = short_status_letter
            else:
                prefix = ''
            prefix = indent + prefix + '  '

            for item in files:
                if item.path[0] is None:
                    path = item.path[1]
                    kind = item.kind[1]
                else:
                    path = item.path[0]
                    kind = item.kind[0]
                if predicate is not None and not predicate(path):
                    continue
                if not header_shown and not short_status:
                    to_file.write(indent + long_status_name + ':\n')
                    header_shown = True
                to_file.write(prefix)
                show_path(path, kind, item.meta_modified(),
                          default_format, with_file_id_format)
                if show_more is not None:
                    show_more(item)
                if show_ids and getattr(item, 'file_id', None):
                    to_file.write(' %s' % item.file_id.decode('utf-8'))
                to_file.write('\n')

    show_list(delta.removed, 'removed', 'D')
    show_list(delta.added, 'added', 'A')
    show_list(delta.missing, 'missing', '!')
    extra_modified = []
    show_list(delta.renamed, 'renamed', 'R', with_file_id_format='%s',
              show_more=show_more_renamed)
    show_list(delta.copied, 'copied', 'C', with_file_id_format='%s',
              show_more=show_more_renamed)
    show_list(delta.kind_changed, 'kind changed', 'K',
              with_file_id_format='%s',
              show_more=show_more_kind_changed)
    show_list(delta.modified + extra_modified, 'modified', 'M')
    if show_unchanged:
        show_list(delta.unchanged, 'unchanged', 'S')

    show_list(delta.unversioned, 'unknown', ' ')



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""VFS operations for the smart server.

This module defines the smart server methods that are low-level file operations
-- i.e. methods that operate directly on files and directories, rather than
higher-level concepts like branches and revisions.

These methods, plus 'hello' and 'get_bundle', are version 1 of the smart server
protocol, as implemented in bzr 0.11 and later.
"""

import os

from ... import urlutils
from . import request


def _deserialise_optional_mode(mode):
    # XXX: FIXME this should be on the protocol object.  Later protocol versions
    # might serialise modes differently.
    if mode == b'':
        return None
    else:
        return int(mode)


def vfs_enabled():
    """Is the VFS enabled ?

    the VFS is disabled when the BRZ_NO_SMART_VFS environment variable is set.

    :return: ``True`` if it is enabled.
    """
    return 'BRZ_NO_SMART_VFS' not in os.environ


class VfsRequest(request.SmartServerRequest):
    """Base class for VFS requests.

    VFS requests are disabled if vfs_enabled() returns False.
    """

    def _check_enabled(self):
        if not vfs_enabled():
            raise request.DisabledMethod(self.__class__.__name__)

    def translate_client_path(self, relpath):
        # VFS requests are made with escaped paths so the escaping done in
        # SmartServerRequest.translate_client_path leads to double escaping.
        # Remove it here -- the fact that the result is still escaped means
        # that the str() will not fail on valid input.
        x = request.SmartServerRequest.translate_client_path(self, relpath)
        return str(urlutils.unescape(x))


class HasRequest(VfsRequest):

    def do(self, relpath):
        relpath = self.translate_client_path(relpath)
        r = self._backing_transport.has(relpath) and b'yes' or b'no'
        return request.SuccessfulSmartServerResponse((r,))


class GetRequest(VfsRequest):

    def do(self, relpath):
        relpath = self.translate_client_path(relpath)
        backing_bytes = self._backing_transport.get_bytes(relpath)
        return request.SuccessfulSmartServerResponse((b'ok',), backing_bytes)


class AppendRequest(VfsRequest):

    def do(self, relpath, mode):
        relpath = self.translate_client_path(relpath)
        self._relpath = relpath
        self._mode = _deserialise_optional_mode(mode)

    def do_body(self, body_bytes):
        old_length = self._backing_transport.append_bytes(
            self._relpath, body_bytes, self._mode)
        return request.SuccessfulSmartServerResponse((b'appended', str(old_length).encode('ascii')))


class DeleteRequest(VfsRequest):

    def do(self, relpath):
        relpath = self.translate_client_path(relpath)
        self._backing_transport.delete(relpath)
        return request.SuccessfulSmartServerResponse((b'ok', ))


class IterFilesRecursiveRequest(VfsRequest):

    def do(self, relpath):
        if not relpath.endswith(b'/'):
            relpath += b'/'
        relpath = self.translate_client_path(relpath)
        transport = self._backing_transport.clone(relpath)
        filenames = transport.iter_files_recursive()
        return request.SuccessfulSmartServerResponse((b'names',) + tuple(filenames))


class ListDirRequest(VfsRequest):

    def do(self, relpath):
        if not relpath.endswith(b'/'):
            relpath += b'/'
        relpath = self.translate_client_path(relpath)
        filenames = self._backing_transport.list_dir(relpath)
        return request.SuccessfulSmartServerResponse((b'names',) + tuple([filename.encode('utf-8') for filename in filenames]))


class MkdirRequest(VfsRequest):

    def do(self, relpath, mode):
        relpath = self.translate_client_path(relpath)
        self._backing_transport.mkdir(relpath,
                                      _deserialise_optional_mode(mode))
        return request.SuccessfulSmartServerResponse((b'ok',))


class MoveRequest(VfsRequest):

    def do(self, rel_from, rel_to):
        rel_from = self.translate_client_path(rel_from)
        rel_to = self.translate_client_path(rel_to)
        self._backing_transport.move(rel_from, rel_to)
        return request.SuccessfulSmartServerResponse((b'ok',))


class PutRequest(VfsRequest):

    def do(self, relpath, mode):
        relpath = self.translate_client_path(relpath)
        self._relpath = relpath
        self._mode = _deserialise_optional_mode(mode)

    def do_body(self, body_bytes):
        self._backing_transport.put_bytes(
            self._relpath, body_bytes, self._mode)
        return request.SuccessfulSmartServerResponse((b'ok',))


class PutNonAtomicRequest(VfsRequest):

    def do(self, relpath, mode, create_parent, dir_mode):
        relpath = self.translate_client_path(relpath)
        self._relpath = relpath
        self._dir_mode = _deserialise_optional_mode(dir_mode)
        self._mode = _deserialise_optional_mode(mode)
        # a boolean would be nicer XXX
        self._create_parent = (create_parent == b'T')

    def do_body(self, body_bytes):
        self._backing_transport.put_bytes_non_atomic(self._relpath,
                                                     body_bytes,
                                                     mode=self._mode,
                                                     create_parent_dir=self._create_parent,
                                                     dir_mode=self._dir_mode)
        return request.SuccessfulSmartServerResponse((b'ok',))


class ReadvRequest(VfsRequest):

    def do(self, relpath):
        relpath = self.translate_client_path(relpath)
        self._relpath = relpath

    def do_body(self, body_bytes):
        """accept offsets for a readv request."""
        offsets = self._deserialise_offsets(body_bytes)
        backing_bytes = b''.join(bytes for offset, bytes in
                                 self._backing_transport.readv(self._relpath, offsets))
        return request.SuccessfulSmartServerResponse((b'readv',), backing_bytes)

    def _deserialise_offsets(self, text):
        # XXX: FIXME this should be on the protocol object.
        offsets = []
        for line in text.split(b'\n'):
            if not line:
                continue
            start, length = line.split(b',')
            offsets.append((int(start), int(length)))
        return offsets


class RenameRequest(VfsRequest):

    def do(self, rel_from, rel_to):
        rel_from = self.translate_client_path(rel_from)
        rel_to = self.translate_client_path(rel_to)
        self._backing_transport.rename(rel_from, rel_to)
        return request.SuccessfulSmartServerResponse((b'ok', ))


class RmdirRequest(VfsRequest):

    def do(self, relpath):
        relpath = self.translate_client_path(relpath)
        self._backing_transport.rmdir(relpath)
        return request.SuccessfulSmartServerResponse((b'ok', ))


class StatRequest(VfsRequest):

    def do(self, relpath):
        if not relpath.endswith(b'/'):
            relpath += b'/'
        relpath = self.translate_client_path(relpath)
        stat = self._backing_transport.stat(relpath)
        return request.SuccessfulSmartServerResponse(
            (b'stat', str(stat.st_size).encode('ascii'), oct(stat.st_mode).encode('ascii')))
